from .base import NeuralNetworkModule, dynamic_module_wrapper, static_module_wrapper

from .resnet import ResNet

from . import base
from . import resnet

__all__ = [
    "NeuralNetworkModule",
    "dynamic_module_wrapper",
    "static_module_wrapper",
    "ResNet",
    "base",
    "resnet",
]



from .a2c import *


class PPO(A2C):
    """
    PPO framework.
    """

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple, Tuple] = (),
        lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
        batch_size: int = 100,
        actor_update_times: int = 5,
        critic_update_times: int = 10,
        actor_learning_rate: float = 0.001,
        critic_learning_rate: float = 0.001,
        entropy_weight: float = None,
        value_weight: float = 0.5,
        surrogate_loss_clip: float = 0.2,
        gradient_max: float = np.inf,
        gae_lambda: float = 1.0,
        discount: float = 0.99,
        normalize_advantage: bool = True,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        **__
    ):
        """
        See Also:
            :class:`.A2C`

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            actor_update_times: Times to update actor in ``update()``.
            critic_update_times: Times to update critic in ``update()``.
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            surrogate_loss_clip: Surrogate loss clipping parameter in PPO.
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
        """
        super().__init__(
            actor,
            critic,
            optimizer,
            criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            actor_update_times=actor_update_times,
            critic_update_times=critic_update_times,
            actor_learning_rate=actor_learning_rate,
            critic_learning_rate=critic_learning_rate,
            entropy_weight=entropy_weight,
            value_weight=value_weight,
            gradient_max=gradient_max,
            gae_lambda=gae_lambda,
            discount=discount,
            normalize_advantage=normalize_advantage,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=replay_buffer,
            visualize=visualize,
            visualize_dir=visualize_dir,
        )
        self.surr_clip = surrogate_loss_clip

    def update(
        self, update_value=True, update_policy=True, concatenate_samples=True, **__
    ):
        # DOC INHERITED
        sum_act_policy_loss = 0
        sum_value_loss = 0

        # create a temporary copy of the not-updated actor
        tmp_actor = deepcopy(self.actor)

        self.actor.train()
        self.critic.train()

        for _ in range(self.actor_update_times):
            # sample a batch
            batch_size, (state, action, advantage) = self.replay_buffer.sample_batch(
                self.batch_size,
                sample_method="random_unique",
                concatenate=concatenate_samples,
                sample_attrs=["state", "action", "gae"],
                additional_concat_attrs=["gae"],
            )

            # normalize advantage
            if self.normalize_advantage:
                advantage = (advantage - advantage.mean()) / (advantage.std() + 1e-6)

            # Infer original action log probability
            # TODO:
            #  This temporary fix is not efficient, maybe requires
            #  PPO store API alternation.
            with t.no_grad():
                self.actor, tmp_actor = tmp_actor, self.actor
                __, action_log_prob, *_ = self._eval_act(state, action)
                self.actor, tmp_actor = tmp_actor, self.actor
                action_log_prob = action_log_prob.view(batch_size, 1)

            if self.entropy_weight is not None:
                __, new_action_log_prob, new_action_entropy, *_ = self._eval_act(
                    state, action
                )
            else:
                __, new_action_log_prob, *_ = self._eval_act(state, action)
                new_action_entropy = None

            new_action_log_prob = new_action_log_prob.view(batch_size, 1)

            # calculate surrogate loss
            # The function of this process is ignoring actions that are not
            # likely to be produced in current actor policy distribution,
            # Because in each update, the old policy distribution diverges
            # from the current distribution more and more.
            sim_ratio = t.exp(new_action_log_prob - action_log_prob)
            advantage = advantage.type_as(sim_ratio)
            surr_loss_1 = sim_ratio * advantage
            surr_loss_2 = (
                t.clamp(sim_ratio, 1 - self.surr_clip, 1 + self.surr_clip) * advantage
            )

            # calculate policy loss using surrogate loss
            act_policy_loss = -t.min(surr_loss_1, surr_loss_2)

            if new_action_entropy is not None:
                act_policy_loss += self.entropy_weight * new_action_entropy.mean()

            act_policy_loss = act_policy_loss.mean()

            if self.visualize:
                self.visualize_model(act_policy_loss, "actor", self.visualize_dir)

            # Update actor network
            if update_policy:
                self.actor.zero_grad()
                self._backward(act_policy_loss)
                nn.utils.clip_grad_norm_(self.actor.parameters(), self.gradient_max)
                self.actor_optim.step()
            sum_act_policy_loss += act_policy_loss.item()

        for _ in range(self.critic_update_times):
            # sample a batch
            batch_size, (state, target_value) = self.replay_buffer.sample_batch(
                self.batch_size,
                sample_method="random_unique",
                concatenate=concatenate_samples,
                sample_attrs=["state", "value"],
                additional_concat_attrs=["value"],
            )
            # calculate value loss
            value = self._criticize(state)
            value_loss = (
                self.criterion(target_value.type_as(value), value) * self.value_weight
            )

            if self.visualize:
                self.visualize_model(value_loss, "critic", self.visualize_dir)

            # Update critic network
            if update_value:
                self.critic.zero_grad()
                self._backward(value_loss)
                nn.utils.clip_grad_norm_(self.critic.parameters(), self.gradient_max)
                self.critic_optim.step()
            sum_value_loss += value_loss.item()

        self.replay_buffer.clear()
        self.actor.eval()
        self.critic.eval()
        return (
            -sum_act_policy_loss / self.actor_update_times,
            sum_value_loss / self.critic_update_times,
        )

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        config = A2C.generate_config(config)
        config["frame"] = "PPO"
        config["frame_config"]["surrogate_loss_clip"] = 0.2
        return config



import torch as t
import numpy as np
import torch.nn as nn
from torch.nn.utils.weight_norm import weight_norm

from .base import NeuralNetworkModule


def conv1x1(in_planes, out_planes, stride=1):
    """
    Create a 1x1 2d convolution block
    """
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


def conv3x3(in_planes, out_planes, stride=1):
    """
    Create a 3x3 2d convolution block
    """
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False
    )


def conv5x5(in_planes, out_planes, stride=2):
    """
    Create a 5x5 2d convolution block
    """
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=5, stride=stride, padding=2, bias=False
    )


def none_norm(*_, **__):
    return nn.Sequential()


def cfg(depth, norm="none"):
    depth_lst = [18, 34, 50, 101, 152]
    if depth not in depth_lst:
        raise ValueError(
            "Error : Resnet depth should be either " "18, 34, 50, 101, 152"
        )
    if norm == "batch":
        cfg_dict = {
            "18": (BasicBlock, [2, 2, 2, 2], {}),
            "34": (BasicBlock, [3, 4, 6, 3], {}),
            "50": (Bottleneck, [3, 4, 6, 3], {}),
            "101": (Bottleneck, [3, 4, 23, 3], {}),
            "152": (Bottleneck, [3, 8, 36, 3], {}),
        }
    elif norm == "weight":
        cfg_dict = {
            "18": (BasicBlockWN, [2, 2, 2, 2], {}),
            "34": (BasicBlockWN, [3, 4, 6, 3], {}),
            "50": (BottleneckWN, [3, 4, 6, 3], {}),
            "101": (BottleneckWN, [3, 4, 23, 3], {}),
            "152": (BottleneckWN, [3, 8, 36, 3], {}),
        }
    elif norm == "none":
        cfg_dict = {
            "18": (BasicBlock, [2, 2, 2, 2], {"norm_layer": none_norm}),
            "34": (BasicBlock, [3, 4, 6, 3], {"norm_layer": none_norm}),
            "50": (Bottleneck, [3, 4, 6, 3], {"norm_layer": none_norm}),
            "101": (Bottleneck, [3, 4, 23, 3], {"norm_layer": none_norm}),
            "152": (Bottleneck, [3, 8, 36, 3], {"norm_layer": none_norm}),
        }
    else:
        raise ValueError(f'Invalid normalization method: "{norm}"')
    return cfg_dict[str(depth)]


class BasicBlock(NeuralNetworkModule):
    # Expansion parameter, output will have "expansion * in_planes" depth.
    expansion = 1

    def __init__(self, in_planes, out_planes, stride=1, norm_layer=None, **__):
        """
        Create a basic block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        super().__init__()
        self.conv1 = conv3x3(in_planes, out_planes, stride)
        self.bn1 = norm_layer(out_planes)
        self.conv2 = conv3x3(out_planes, self.expansion * out_planes)
        self.bn2 = norm_layer(self.expansion * out_planes)
        # Create a shortcut from input to output.
        # An empty sequential structure means no transformation
        # is made on input X.
        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        # A convolution is needed if we cannot directly add input X to output
        # BatchNorm2d produces NaN gradient?
        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                conv1x1(in_planes, self.expansion * out_planes, stride),
                norm_layer(self.expansion * out_planes),
            )

    def forward(self, x):
        out = t.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = t.relu(out)
        return out


class Bottleneck(NeuralNetworkModule):
    # Expansion parameter, output will have "expansion * in_planes" depth.
    expansion = 4

    def __init__(self, in_planes, out_planes, stride=1, norm_layer=None, **__):
        """
        Create a bottleneck block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        super().__init__()
        self.conv1 = conv1x1(in_planes, out_planes)
        self.conv2 = conv3x3(out_planes, out_planes, stride=stride)
        self.conv3 = conv1x1(out_planes, self.expansion * out_planes)
        self.bn1 = norm_layer(out_planes)
        self.bn2 = norm_layer(out_planes)
        self.bn3 = norm_layer(self.expansion * out_planes)

        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                conv1x1(in_planes, self.expansion * out_planes, stride),
                norm_layer(self.expansion * out_planes),
            )

    def forward(self, x):
        out = t.relu(self.bn1(self.conv1(x)))
        out = t.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = t.relu(out)
        return out


class BasicBlockWN(NeuralNetworkModule):
    """
    Basic block with weight normalization
    """

    # Expansion parameter, output will have "expansion * in_planes" depth.
    expansion = 1

    def __init__(self, in_planes, out_planes, stride=1, **__):
        """
        Create a basic block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        super().__init__()
        self.conv1 = weight_norm(conv3x3(in_planes, out_planes, stride))
        self.conv2 = weight_norm(conv3x3(out_planes, self.expansion * out_planes))
        # Create a shortcut from input to output.
        # An empty sequential structure means no transformation
        # is made on input X.
        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        # A convolution is needed if we cannot directly add input X to output.
        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                weight_norm(conv1x1(in_planes, self.expansion * out_planes, stride)),
            )

    def forward(self, x):
        out = t.relu(self.conv1(x))
        out = self.conv2(out)
        out += self.shortcut(x)
        out = t.relu(out)
        return out


class BottleneckWN(NeuralNetworkModule):
    """
    Bottleneck block with weight normalization
    """

    # expansion parameter, output will have "expansion * in_planes" depth
    expansion = 4

    def __init__(self, in_planes, out_planes, stride=1, **__):
        """
        Create a bottleneck block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        super().__init__()
        self.conv1 = weight_norm(
            nn.Conv2d(in_planes, out_planes, kernel_size=1, bias=False)
        )
        self.conv2 = weight_norm(
            nn.Conv2d(
                out_planes,
                out_planes,
                kernel_size=3,
                stride=stride,
                padding=1,
                bias=False,
            )
        )
        self.conv3 = weight_norm(
            nn.Conv2d(
                out_planes, self.expansion * out_planes, kernel_size=1, bias=False
            )
        )

        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                weight_norm(conv1x1(in_planes, self.expansion * out_planes, stride)),
            )

    def forward(self, x):
        out = t.relu(self.conv1(x))
        out = t.relu(self.conv2(out))
        out = self.conv3(out)
        out += self.shortcut(x)
        out = t.relu(out)

        return out


class ResNet(NeuralNetworkModule):
    def __init__(
        self,
        in_planes: int,
        depth: int,
        out_planes: int,
        out_pool_size=(1, 1),
        norm="none",
    ):
        """
        Create a resnet of specified depth.

        Args:
            in_planes:  Number of input planes.
            depth: Depth of resnet. Could be one of ``18, 34, 50, 101, 152``.
            out_planes: Number of output planes.
            out_pool_size: Size of pooling output
            norm: Normalization method, could be one of "none", "batch" or
                "weight".
        """
        super().__init__()
        self.norm = norm
        self.depth = depth
        self.in_planes = in_planes
        self.out_planes = out_planes
        self.out_pool_size = out_pool_size

        self._cur_in_planes = 64

        block, num_blocks, kw = cfg(depth, norm)

        self.conv1 = nn.Conv2d(
            in_planes, 64, kernel_size=7, stride=2, padding=3, bias=False
        )
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        if norm == "batch":
            self.bn1 = nn.BatchNorm2d(64)
            self.layer1 = self._make_layer(block, 64, num_blocks[0], 1, kw)
            self.layer2 = self._make_layer(block, 128, num_blocks[1], 2, kw)
            self.layer3 = self._make_layer(block, 256, num_blocks[2], 2, kw)
            self.layer4 = self._make_layer(block, 512, num_blocks[3], 2, kw)
            self.base = nn.Sequential(
                self.conv1,
                self.bn1,
                nn.ReLU(),
                self.maxpool,
                self.layer1,
                self.layer2,
                self.layer3,
                self.layer4,
            )
        else:
            self.layer1 = self._make_layer(block, 64, num_blocks[0], 1, kw)
            self.layer2 = self._make_layer(block, 128, num_blocks[1], 2, kw)
            self.layer3 = self._make_layer(block, 256, num_blocks[2], 2, kw)
            self.layer4 = self._make_layer(block, 512, num_blocks[3], 2, kw)
            self.base = nn.Sequential(
                self.conv1,
                nn.ReLU(),
                self.maxpool,
                self.layer1,
                self.layer2,
                self.layer3,
                self.layer4,
            )
        self.fc = nn.Linear(512 * out_pool_size[0] * out_pool_size[1], out_planes)

        self.set_input_module(self.conv1)

    def _make_layer(self, block, planes, num_blocks, stride, kwargs):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []

        for stride in strides:
            layers.append(block(self._cur_in_planes, planes, stride, **kwargs))
            self._cur_in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        assert x.shape[2] == 224 and x.shape[3] == 224
        x = self.base(x)
        kernel_size = (
            np.int(np.floor(x.size(2) / self.out_pool_size[0])),
            np.int(np.floor(x.size(3) / self.out_pool_size[1])),
        )
        x = nn.functional.avg_pool2d(x, kernel_size)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x



from abc import ABC, abstractmethod
from collections import OrderedDict
from threading import Lock
from copy import deepcopy
from ..distributed import RpcGroup, debug_with_process


class OrderedServerBase(ABC):  # pragma: no cover
    """
    Descendent classes of OrderedServer does not have to guarantee strong
    consistency, that is, even if :meth:`.OrderedServerBase.push_service``
    has returned True, there are possibilities that these acknowledged
    push are discarded.
    """

    @abstractmethod
    def push(self, key, value, version, prev_version):
        """
        Push a new ``version`` of ``value`` in ``key`` to the ordered server.

        Note:
            If ``version = prev_version`` then there is no order guarantee. But
            you may exploit this feature.

        Args:
            key: Key.
            value: value.
            version: New version.
            prev_version: Previous version.

        Returns:
            ``True`` if success, and ``False`` if not.
        """
        pass

    @abstractmethod
    def pull(self, key, version=None):
        """
        Pull a value with the specified ``version`` in ``key``.

        Args:
            key: Key.
            version: Target version, if ``None``, then the newest version
                of value of key will be pulled.

        Returns:
            ``None`` if version is not found, auto-deleted, or key is not found,
            otherwise returns value with the specified ``version``
            in ``key``, and then ``version``
        """
        pass


class OrderedServerSimple(OrderedServerBase):
    def __init__(self, server_name: str, group: RpcGroup):
        self._push_service = server_name + "/_push_service"
        self._pull_service = server_name + "/_pull_service"
        self.group = group

    def push(self, key, value, version, prev_version):
        # DOC INHERITED
        debug_with_process(
            f"calling push service {self._push_service} "
            f"on group [{self.group.group_name}]"
        )
        return self.group.registered_sync(
            self._push_service, args=(key, value, version, prev_version)
        )

    def pull(self, key, version=None):
        # DOC INHERITED
        debug_with_process(
            f"calling pull service {self._push_service} "
            f"on group [{self.group.group_name}]"
        )
        return self.group.registered_sync(self._pull_service, args=(key, version))


class OrderedServerSimpleImpl:
    """
    A simple key-value server, with strict ordered update
    """

    def __init__(self, server_name: str, group: RpcGroup, version_depth: int = 1, **__):
        """
        This init function must be only invoked on the runner process,
        and the runner process must be a member process of ``group``.

        Args:
            server_name: Name of this server, used to registered
                the server as a paired class of ``group``.
            group: Rpc group where server locates.
            server_runner: Name of the process serving the ordered server.
                By default is the first member of the group.
            version_depth: Storage depth of old versions of the same
                key. If ``depth = 1``, then only the newest version
                of the key will be saved.
        """
        assert group.is_member()
        assert version_depth > 0 and isinstance(version_depth, int)

        self.server_name = server_name
        self.group = group
        self.storage = {}
        self.lock = Lock()
        self.version_depth = version_depth
        # pair an accessor to group
        self.group.pair(server_name, OrderedServerSimple(self.server_name, self.group))
        self.group.register(server_name + "/_push_service", self._push_service)
        self.group.register(server_name + "/_pull_service", self._pull_service)

    def _push_service(self, key, value, version, prev_version):
        success = False
        with self.lock:
            if key in self.storage:
                ref = self.storage[key]
                # Check previous version consistency.
                if next(reversed(ref)) == prev_version:
                    ref[version] = value
                    success = True
                if len(ref) > self.version_depth + 1:
                    ref.popitem(last=False)
            else:
                # Create a new key.
                ref = self.storage[key] = OrderedDict()
                ref[version] = value
                success = True
        return success

    def _pull_service(self, key, version=None):
        result = None
        with self.lock:
            if key in self.storage:
                ref = self.storage[key]
                # Try to find the target version.
                if version is not None and version in ref:
                    result = (deepcopy(ref[version]), version)
                # Find the newest version.
                elif version is None:
                    version = next(reversed(ref))
                    result = (deepcopy(ref[version]), version)
        return result



from copy import deepcopy
from typing import Dict, List, Any, Union
from ..config import fill_default, is_algorithm_distributed
from ..pl_logger import LocalMediaLogger
from ..dataset import DatasetResult, RLDataset, log_video, determine_precision
from ..launcher import Launcher, DistributedLauncher
from machin.frame.algorithms import *
from machin.env.utils.openai_gym import disable_view_window
from machin.utils.conf import Config
from machin.utils.save_env import SaveEnv
from pytorch_lightning.callbacks import Callback, ModelCheckpoint, EarlyStopping
from pytorch_lightning.loggers import TensorBoardLogger
from gym.spaces import Box, Discrete, MultiBinary, MultiDiscrete
import gym
import torch as t
import pytorch_lightning as pl


disable_view_window()


def _is_simple_space(space):
    return type(space) in (Box, Discrete, MultiBinary, MultiDiscrete)


def _is_discrete_space(space):
    return type(space) in (Discrete, MultiDiscrete)


def _is_continuous_space(space):
    return type(space) in (Box, MultiBinary)


class RLGymDiscActDataset(RLDataset):
    """
    This dataset is using a discrete action openai-gym environment.

    Notes:
        The forward method of Q networks, actor networks must accept arguments
        of default names like "action", "state", and not custom names like
        "some_action", "some_state".

        All your networks should not have custom outputs after default ones
        like action, action_log_prob, etc.

    Notes:
        The environment should accept an python int number in each step.

        The environment should output a simple observation space, dict and
        tuple are not supported. The first dimension size should be 1 as it will
        be used as the batch size, if not, it will be automatically added.

        The environment should have a finite number of acting steps.

    Args:
        frame: Algorithm framework.
        env: Gym environment instance.
        act_kwargs: Additional keyword arguments passed to act functions
            of different frameworks.
    """

    early_stopping_monitor = "total_reward"

    def __init__(
        self,
        frame,
        env,
        render_every_episode: int = 100,
        act_kwargs: Dict[str, Any] = None,
    ):
        super().__init__()
        self.frame = frame
        self.env = env
        self.render_every_episode = render_every_episode
        self.act_kwargs = act_kwargs or {}
        self._precision = determine_precision(
            [getattr(frame, m) for m in frame._is_top]
        )
        self.counter = 0
        assert type(env.action_space) == Discrete
        assert _is_simple_space(env.observation_space)

    def __next__(self):
        result = DatasetResult()
        terminal = False
        total_reward = 0
        state = t.tensor(self.env.reset(), dtype=self._precision)
        state = state.flatten().unsqueeze(0)

        # manual sync and then disable syncing if framework is distributed.
        getattr(self.frame, "manual_sync", lambda: None)()
        getattr(self.frame, "set_sync", lambda x: None)(False)

        rendering = []
        while not terminal:
            if self.counter % self.render_every_episode == 0:
                rendering.append(self.env.render(mode="rgb_array"))

            with t.no_grad():
                old_state = state
                # agent model inference
                if type(self.frame) in (A2C, PPO, SAC, A3C, IMPALA):
                    action = self.frame.act({"state": old_state}, **self.act_kwargs)[0]
                elif type(self.frame) in (DQN, DQNPer, DQNApex, RAINBOW):
                    action = self.frame.act_discrete_with_noise(
                        {"state": old_state}, **self.act_kwargs
                    )
                elif type(self.frame) in (DDPG, DDPGPer, HDDPG, TD3, DDPGApex):
                    action, probs = self.frame.act_discrete_with_noise(
                        {"state": old_state}, **self.act_kwargs
                    )
                elif type(self.frame) in (ARS,):
                    action = self.frame.act({"state": old_state}, **self.act_kwargs)
                else:
                    raise RuntimeError(f"Unsupported framework: {type(self.frame)}")
                state, reward, terminal, info = self.env.step(action.item())
                state = t.tensor(state, dtype=self._precision)
                state = state.flatten().unsqueeze(0)
                reward = float(reward)
                total_reward += reward
                if type(self.frame) in (DDPG, DDPGPer, HDDPG, TD3, DDPGApex):
                    action = probs
                result.add_observation(
                    {
                        "state": {"state": old_state},
                        "action": {"action": action},
                        "next_state": {"state": state},
                        "reward": reward,
                        "terminal": terminal,
                    }
                )
                result.add_log(info)

        if len(rendering) > 0:
            result.add_log({"video": (rendering, log_video)})
        result.add_log({"total_reward": total_reward})

        getattr(self.frame, "set_sync", lambda x: None)(True)
        self.counter += 1
        return result


class RLGymContActDataset(RLDataset):
    """
    This dataset is using a contiguous action openai-gym environment.

    Notes:
        The forward method of Q networks, actor networks must accept arguments
        of default names like "action", "state", and not custom names like
        "some_action", "some_state".

        All your networks should not have custom outputs after default ones
        like action, action_log_prob, etc.

    Notes:
        The environment should accept a numpy float array in each step.

        The environment should output a simple observation space, dict and
        tuple are not supported. The first dimension size should be 1 as it will
        be used as the batch size, if not, it will be automatically added.

        The environment should have a finite number of acting steps.

    Args:
        frame: Algorithm framework.
        env: Gym environment instance.
        act_kwargs: Additional keyword arguments passed to act functions
            of different frameworks.
    """

    early_stopping_monitor = "total_reward"

    def __init__(
        self,
        frame,
        env,
        render_every_episode: int = 100,
        act_kwargs: Dict[str, Any] = None,
    ):
        super().__init__()
        self.frame = frame
        self.env = env
        self.render_every_episode = render_every_episode
        self.act_kwargs = act_kwargs or {}
        self._precision = determine_precision(
            [getattr(frame, m) for m in frame._is_top]
        )
        self.counter = 0
        assert type(env.action_space) == Box
        assert _is_simple_space(env.observation_space)

    def __next__(self):
        result = DatasetResult()
        terminal = False
        total_reward = 0
        state = t.tensor(self.env.reset(), dtype=self._precision)
        state = state.flatten().unsqueeze(0)

        # manual sync and then disable syncing if framework is distributed.
        getattr(self.frame, "manual_sync", lambda: None)()
        getattr(self.frame, "set_sync", lambda x: None)(False)

        rendering = []

        while not terminal:
            if self.counter % self.render_every_episode == 0:
                rendering.append(self.env.render(mode="rgb_array"))

            with t.no_grad():
                old_state = state
                # agent model inference
                if type(self.frame) in (A2C, PPO, SAC, A3C, IMPALA):
                    action = self.frame.act({"state": old_state}, **self.act_kwargs)[0]
                elif type(self.frame) in (DDPG, DDPGPer, HDDPG, TD3, DDPGApex):
                    action = self.frame.act_with_noise(
                        {"state": old_state}, **self.act_kwargs
                    )[0]
                elif type(self.frame) in (ARS,):
                    action = self.frame.act({"state": old_state}, **self.act_kwargs)
                else:
                    raise RuntimeError(f"Unsupported framework: {type(self.frame)}")
                state, reward, terminal, info = self.env.step(
                    action.detach().cpu().numpy()
                )
                state = t.tensor(state, dtype=self._precision)
                state = state.flatten().unsqueeze(0)
                reward = float(reward)
                total_reward += reward
                result.add_observation(
                    {
                        "state": {"state": old_state},
                        "action": {"action": action},
                        "next_state": {"state": state},
                        "reward": reward,
                        "terminal": terminal,
                    }
                )
                result.add_log(info)

        if len(rendering) > 0:
            result.add_log({"video": (rendering, log_video)})
        result.add_log({"total_reward": total_reward})

        getattr(self.frame, "set_sync", lambda x: None)(True)
        self.counter += 1
        return result


def gym_env_dataset_creator(frame, env_config):
    env = gym.make(env_config["env_name"])
    if _is_discrete_space(env.action_space):
        return RLGymDiscActDataset(
            frame,
            env,
            render_every_episode=env_config["render_every_episode"],
            act_kwargs=env_config["act_kwargs"],
        )
    elif _is_continuous_space(env.action_space):
        return RLGymContActDataset(
            frame,
            env,
            render_every_episode=env_config["render_every_episode"],
            act_kwargs=env_config["act_kwargs"],
        )
    else:
        raise ValueError(
            f"Gym environment {env_config['env_name']} has action space "
            f"of type {env.action_space}, which is not supported."
        )


def generate_env_config(
    env_name: str = None, config: Union[Dict[str, Any], Config] = None
):
    """
    Generate example OpenAI gym config.
    """
    config = deepcopy(config) or {}
    return fill_default(
        {
            "env": "openai_gym",
            "train_env_config": {
                "env_name": env_name or "CartPole-v1",
                "render_every_episode": 100,
                "act_kwargs": {},
            },
            "test_env_config": {
                "env_name": env_name or "CartPole-v1",
                "render_every_episode": 100,
                "act_kwargs": {},
            },
        },
        config,
    )


def launch(config: Union[Dict[str, Any], Config], pl_callbacks: List[Callback] = None):
    """
    Args:
        config: All configs needed to launch a gym environment and initialize
            the algorithm framework.
        pl_callbacks: Additional callbacks used to modify training behavior.

    Returns:

    """
    pl_callbacks = pl_callbacks or []

    # disable time formatting so all processes will use the same directory
    # delay directory creation
    s_env = SaveEnv(config.get("root_dir", None) or "./trial", time_format="")

    checkpoint_callback = ModelCheckpoint(
        dirpath=s_env.get_trial_model_dir(),
        filename="{epoch:02d}-{total_reward:.2f}",
        save_top_k=1,
        monitor="total_reward",
        mode="max",
        period=1,
        verbose=True,
    )
    early_stopping = EarlyStopping(
        monitor="total_reward", mode="max", patience=config["early_stopping_patience"],
    )
    t_logger = TensorBoardLogger(s_env.get_trial_train_log_dir())
    lm_logger = LocalMediaLogger(
        s_env.get_trial_image_dir(), s_env.get_trial_image_dir()
    )

    is_distributed = is_algorithm_distributed(config)
    trainer = pl.Trainer(
        gpus=config.get("gpus", None),
        num_processes=config.get("num_processes", 1),
        num_nodes=config.get("num_nodes", 1),
        callbacks=[checkpoint_callback, early_stopping] + pl_callbacks,
        logger=[t_logger, lm_logger],
        limit_train_batches=config["episode_per_epoch"],
        max_steps=config["max_episodes"],
        accelerator="ddp" if is_distributed else None,
    )
    if is_distributed:
        model = DistributedLauncher(config, gym_env_dataset_creator)
    else:
        model = Launcher(config, gym_env_dataset_creator)
    trainer.fit(model)



from PIL import Image
from typing import List
from machin.parallel import get_context
import os
import numpy as np
import moviepy.editor as mpy
import matplotlib.pyplot as plt


def show_image(
    image: np.ndarray,
    show_normalized: bool = True,
    pause_time: float = 0.01,
    title: str = "",
):
    """
    Use matplotlib to show a single image. You may repeatedly call this method
    with the same ``title`` argument to show a video or a dynamically changing
    image.

    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with ``dtype``
            = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        show_normalized: Show normalized image alongside the original one.
        pause_time: Pause time between displaying current image and the next
            one.
        title: Title of the display window.
    """
    if np.issubdtype(image.dtype, np.integer):
        image = image.astype(np.floating) / 255

    fig = plt.figure(title, clear=True)
    fig.canvas.set_window_title(title)
    if show_normalized:
        ax = fig.add_subplot("121")
        ax.set_facecolor((0.0, 0.0, 0.0))
        ax.imshow(image, vmin=np.min(image), vmax=np.max(image))

        ax2 = fig.add_subplot("122")
        ax2.set_facecolor((0.0, 0.0, 0.0))
        pix_range = (np.max(image) - np.min(image)) + 1e-6
        ax2.imshow((image - np.min(image)) / pix_range, vmin=0, vmax=1)
        plt.pause(pause_time)

    else:
        ax = fig.add_subplot("111")
        ax.set_facecolor((0.0, 0.0, 0.0))
        ax.imshow(image, vmin=np.min(image), vmax=np.max(image))
        plt.pause(pause_time)


def create_video(
    frames: List[np.ndarray],
    path: str,
    filename: str,
    extension: str = ".gif",
    fps: int = 15,
):
    """
    Args:
        frames: A list of numpy arrays of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the video.
        filename: File name.
        extension: File extension.
        fps: frames per second.
    """
    if frames:
        for f in range(len(frames)):
            if np.issubdtype(frames[f].dtype, np.integer):
                frames[f] = frames[f].astype(np.uint8)
            elif np.issubdtype(frames[f].dtype, np.floating):
                frames[f] = (frames[f] * 255).astype(np.uint8)
            if frames[f].ndim == 2:
                # consider as a grey scale image
                frames[f] = np.repeat(frames[f][:, :, np.newaxis], 3, axis=2)

        clip = mpy.ImageSequenceClip(frames, fps=fps)
        if extension.lower() == ".gif":
            clip.write_gif(
                os.path.join(path, filename + extension),
                fps=fps,
                verbose=False,
                logger=None,
            )
        else:
            clip.write_videofile(
                os.path.join(path, filename + extension),
                fps=fps,
                verbose=False,
                logger=None,
            )
        clip.close()


def create_video_subproc(
    frames: List[np.ndarray],
    path: str,
    filename: str,
    extension: str = ".gif",
    fps: int = 15,
    daemon: bool = True,
):
    """
    Create video with a subprocess, since it takes a lot of time for ``moviepy``
    to encode the video file.

    See Also:
         :func:`.create_video`

    Note:
        if ``daemon`` is true, then this function cannot be used in a
        daemonic subprocess.

    Args:
        frames: A list of numpy arrays of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the video.
        filename: File name.
        extension: File extension.
        fps: frames per second.
        daemon: Whether launching the saving process as a daemonic process.

    Returns:
        A wait function, once called, block until creation has finished.
    """

    def wait():
        pass

    if frames:
        p = get_context("spawn").Process(
            target=create_video, args=(frames, path, filename, extension, fps)
        )
        p.daemon = daemon
        p.start()

        def wait():
            p.join()

    return wait


def numpy_array_to_pil_image(image: np.ndarray):
    if np.issubdtype(image.dtype, np.integer):
        image = image.astype(np.uint8)
    elif np.issubdtype(image.dtype, np.floating):
        image = (image * 255).astype(np.uint8)
    if image.ndim == 2:
        # consider as a grey scale image
        image = np.repeat(image[:, :, np.newaxis], 3, axis=2)
    image = Image.fromarray(image)
    return image


def create_image(image: np.ndarray, path: str, filename: str, extension: str = ".png"):
    """
    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the image.
        filename: File name.
        extension: File extension.
    """
    image = numpy_array_to_pil_image(image)
    image.save(os.path.join(path, filename + extension))


def create_image_subproc(
    image: np.array,
    path: str,
    filename: str,
    extension: str = ".png",
    daemon: bool = True,
):
    """
    Create image with a subprocess.

    See Also:
         :func:`.create_image`

    Note:
        if ``daemon`` is true, then this function cannot be used in a
        daemonic subprocess.

    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the image.
        filename: File name.
        extension: File extension.
        daemon: Whether launching the saving process as a daemonic process.

    Returns:
        A wait function, once called, block until creation has finished.
    """
    p = get_context("spawn").Process(
        target=create_image, args=(image, path, filename, extension)
    )
    p.daemon = daemon
    p.start()

    def wait():
        p.join()

    return wait



import copy
import inspect
import itertools
from random import choice, randint
from .utils import determine_device
from machin.utils.visualize import make_dot
from machin.utils.logging import default_logger
from machin.model.nets.base import static_module_wrapper
from machin.parallel.pool import P2PPool, ThreadPool

# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


class SHMBuffer(Buffer):
    @staticmethod
    def make_tensor_from_batch(batch, device, concatenate):
        # this function is used in post processing, and we will
        # move all cpu tensors to shared memory.
        if concatenate and len(batch) != 0:
            item = batch[0]
            batch_size = len(batch)
            if t.is_tensor(item):
                batch = [it.to(device) for it in batch]
                result = t.cat(batch, dim=0).to(device)
                result.share_memory_()
                return result
            else:
                try:
                    result = t.tensor(batch, device=device).view(batch_size, -1)
                    result.share_memory_()
                    return result
                except Exception:
                    raise ValueError(f"Batch not concatenable: {batch}")
        else:
            for it in batch:
                if t.is_tensor(it):
                    it.share_memory_()
            return batch


class MADDPG(TorchFramework):
    """
    MADDPG is a centralized multi-agent training framework, it alleviates the
    unstable reward problem caused by the disturbance of other agents by
    gathering all agents observations and train a global critic. This global
    critic observes all actions and all states from all agents.
    """

    # Since the number of sub-policies is automatically determined,
    # they are not considered here.
    _is_top = ["all_actor_target", "all_critic_target"]
    _is_restorable = ["all_actor_target", "all_critic_target"]

    def __init__(
        self,
        actors: List[Union[NeuralNetworkModule, nn.Module]],
        actor_targets: List[Union[NeuralNetworkModule, nn.Module]],
        critics: List[Union[NeuralNetworkModule, nn.Module]],
        critic_targets: List[Union[NeuralNetworkModule, nn.Module]],
        optimizer: Callable,
        criterion: Callable,
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[List[Tuple], List[Tuple]] = None,
        lr_scheduler_kwargs: Tuple[List[Dict], List[Dict]] = None,
        critic_visible_actors: List[List[int]] = None,
        sub_policy_num: int = 0,
        batch_size: int = 100,
        update_rate: float = 0.001,
        update_steps: Union[int, None] = None,
        actor_learning_rate: float = 0.0005,
        critic_learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        replay_device: Union[str, t.device] = "cpu",
        replay_buffer: Buffer = None,
        visualize: bool = False,
        visualize_dir: str = "",
        use_jit: bool = True,
        pool_type: str = "thread",
        pool_size: int = None,
        **__,
    ):
        """
        See Also:
            :class:`.DDPG`

        Note:
            In order to parallelize agent inference, a process pool is used
            internally. However, in order to minimize memory copy / CUDA memory
            copy, the location of all of your models must be either "cpu", or
            "cuda" (Using multiple CUDA devices is supported).

        Note:
            MADDPG framework **does not require** all of your actors are
            homogeneous. Each pair of your actors and critcs could be
            heterogeneous.

        Note:
            Suppose you have three pair of actors and critics, with index 0, 1,
            2. If critic 0 can observe the action of actor 0 and 1, critic 1 can
            observe the action of actor 1 and 2, critic 2 can observe the action
            of actor 2 and 0, the ``critic_visible_actors`` should be::

                [[0, 1], [1, 2], [2, 0]]

        Note:
            Learning rate scheduler args and kwargs for each actor and critic,
            the first list is for actors, and the second list is for critics.

        Note:
            This implementation contains:
                - Ensemble Training

            This implementation does not contain:
                - Inferring other agents' policies
                - Mixed continuous/discrete action spaces

        Args:
            actors: Actor network modules.
            actor_targets: Target actor network modules.
            critics: Critic network modules.
            critic_targets: Target critic network modules.
            optimizer: Optimizer used to optimize ``actors`` and ``critics``.
                By default all critics can see outputs of all actors.
            criterion: Criterion used to evaluate the value loss.
            critic_visible_actors: Indexes of visible actors for each critic.
            sub_policy_num: Times to replicate each actor. Equals to
                `ensemble_policy_num - 1`
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:
                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            update_steps: Training step number used to update target networks.
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size for each actor. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer. Will be replicated for actor.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
            use_jit: Whether use torch jit to perform the forward pass
                in parallel instead of using the internal pool. Provides
                significant speed and efficiency advantage, but requires
                actors and critics convertible to TorchScript.
            pool_type: Type of the internal execution pool, either "process"
                or "thread".
            pool_size: Size of the internal execution pool.
        """
        assert pool_type in ("process", "thread")
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.update_steps = update_steps
        self.discount = discount
        self.has_visualized = False
        self.visualize = visualize
        self.visualize_dir = visualize_dir
        self.grad_max = gradient_max
        self.critic_visible_actors = critic_visible_actors or [
            list(range(len(actors)))
        ] * len(actors)
        self._update_counter = 0

        if update_rate is not None and update_steps is not None:
            raise ValueError(
                "You can only specify one target network update"
                " scheme, either by update_rate or update_steps,"
                " but not both."
            )

        # create ensembles of policies
        self.actors = [
            [actor] + [copy.deepcopy(actor) for _ in range(sub_policy_num)]
            for actor in actors
        ]
        self.actor_targets = [
            [actor_target]
            + [copy.deepcopy(actor_target) for _ in range(sub_policy_num)]
            for actor_target in actor_targets
        ]
        self.critics = critics
        self.critic_targets = critic_targets
        self.actor_optims = [
            [optimizer(acc.parameters(), lr=actor_learning_rate) for acc in ac]
            for ac in self.actors
        ]
        self.critic_optims = [
            optimizer(cr.parameters(), lr=critic_learning_rate) for cr in self.critics
        ]
        self.ensemble_size = sub_policy_num + 1
        self.replay_buffers = [
            SHMBuffer(replay_size, replay_device)
            if replay_buffer is None
            else copy.deepcopy(replay_buffer)
            for _ in range(len(actors))
        ]

        # create the pool used to update()
        # check devices of all parameters,
        # determine the pool process starting method.
        device = self._check_parameters_device(
            itertools.chain(*self.actors, self.critics)
        )
        self.device = device

        self.pool_type = pool_type
        if pool_type == "process":
            self.pool = P2PPool(
                processes=pool_size,
                is_recursive=False,
                is_copy_tensor=False,
                share_method=device,
            )
        elif pool_type == "thread":
            self.pool = ThreadPool(processes=pool_size)

        # Create wrapper for target actors and target critics.
        # So their parameters can be saved.
        self.all_actor_target = nn.Module()
        self.all_critic_target = nn.Module()

        for ac, idx in zip(self.actor_targets, range(len(actors))):
            for acc, idxx in zip(ac, range(self.ensemble_size)):
                acc.share_memory()
                self.all_actor_target.add_module(f"actor_{idx}_{idxx}", acc)

        for cr, idx in zip(self.critic_targets, range(len(critics))):
            cr.share_memory()
            self.all_critic_target.add_module(f"critic_{idx}", cr)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            self.pool.starmap(
                hard_update,
                zip(
                    itertools.chain(*self.actors), itertools.chain(*self.actor_targets)
                ),
            )
            self.pool.starmap(hard_update, zip(self.critics, self.critic_targets))

        if lr_scheduler is not None:
            if lr_scheduler_args is None:
                lr_scheduler_args = ([()] * len(actors), [()] * len(critics))
            if lr_scheduler_kwargs is None:
                lr_scheduler_kwargs = ([{}] * len(actors), [{}] * len(critics))
            self.actor_lr_schs = [
                lr_scheduler(acc_opt, *lr_sch_args, *lr_sch_kwargs)
                for ac_opt, lr_sch_args, lr_sch_kwargs in zip(
                    self.actor_optims, lr_scheduler_args[0], lr_scheduler_kwargs[0]
                )
                for acc_opt in ac_opt
            ]
            self.critic_lr_schs = [
                lr_scheduler(cr_opt, *lr_sch_args, *lr_sch_kwargs)
                for cr_opt, lr_sch_args, lr_sch_kwargs in zip(
                    self.critic_optims, lr_scheduler_args[1], lr_scheduler_kwargs[1]
                )
            ]

        self.criterion = criterion

        # make preparations if use jit
        # jit modules will share the same parameter memory with original
        # modules, therefore it is safe to use them together.
        self.use_jit = use_jit
        self.jit_actors = []
        self.jit_actor_targets = []
        if use_jit:
            # only compile actors, since critics will not be
            # launched in parallel
            for ac in self.actors:
                jit_actors = []
                jit_actor_targets = []
                for acc in ac:
                    # exclude "self" by truncating element 0
                    actor_arg_spec = inspect.getfullargspec(acc.forward)
                    jit_actor = t.jit.script(acc)
                    jit_actor.arg_spec = actor_arg_spec
                    jit_actor.model_type = type(acc)
                    jit_actors.append(jit_actor)

                    jit_actor_target = t.jit.script(acc)
                    jit_actor_target.arg_spec = actor_arg_spec
                    jit_actor_target.model_type = type(acc)
                    jit_actor_targets.append(jit_actor_target)
                self.jit_actors.append(jit_actors)
                self.jit_actor_targets.append(jit_actor_targets)

        super().__init__()

    @property
    def optimizers(self):
        return sum(self.actor_optims, self.critic_optims)

    @optimizers.setter
    def optimizers(self, optimizers):
        counter = 0
        for ac in self.actor_optims:
            for id, _acc in enumerate(ac):
                ac[id] = optimizers[counter]
                counter += 1
        for id in range(len(self.critic_optims)):
            self.critic_optims[id] = optimizers[counter]
            counter += 1

    @property
    def lr_schedulers(self):
        if hasattr(self, "actor_lr_schs") and hasattr(self, "critic_lr_schs"):
            return self.actor_lr_schs + self.critic_lr_schs
        return []

    def act(self, states: List[Dict[str, Any]], use_target: bool = False, **__):
        """
        Use all actor networks to produce actions for the current state.
        A random sub-policy from the policy ensemble of each actor will
        be chosen.

        Args:
            states: A list of current states of each actor.
            use_target: Whether use the target network.

        Returns:
            A list of anything returned by your actor. If your actor
            returns multiple values, they will be wrapped in a tuple.
        """
        return [safe_return(act) for act in self._act_api_general(states, use_target)]

    def act_with_noise(
        self,
        states: List[Dict[str, Any]],
        noise_param: Any = (0.0, 1.0),
        ratio: float = 1.0,
        mode: str = "uniform",
        use_target: bool = False,
        **__,
    ):
        """
        Use all actor networks to produce noisy actions for the current state.
        A random sub-policy from the policy ensemble of each actor will
        be chosen.

        See Also:
             :mod:`machin.frame.noise.action_space_noise`

        Args:
            states: A list of current states of each actor.
            noise_param: Noise params.
            ratio: Noise ratio.
            mode: Noise mode. Supported are:
                ``"uniform", "normal", "clipped_normal", "ou"``
            use_target: Whether use the target network.

        Returns:
            A list of noisy actions of shape ``[batch_size, action_dim]``.
        """
        actions = self._act_api_general(states, use_target)
        result = []
        for action, *others in actions:
            if mode == "uniform":
                action = add_uniform_noise_to_action(action, noise_param, ratio)
            elif mode == "normal":
                action = add_normal_noise_to_action(action, noise_param, ratio)
            elif mode == "clipped_normal":
                action = add_clipped_normal_noise_to_action(action, noise_param, ratio)
            elif mode == "ou":
                action = add_ou_noise_to_action(action, noise_param, ratio)
            else:
                raise ValueError("Unknown noise type: " + str(mode))
            if len(others) == 0:
                result.append(action)
            else:
                result.append((action, *others))
        return result

    def act_discrete(self, states: List[Dict[str, Any]], use_target: bool = False):
        """
        Use all actor networks to produce discrete actions for the current
        state.
        A random sub-policy from the policy ensemble of each actor will
        be chosen.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            states: A list of current states of each actor.
            use_target: Whether use the target network.

        Returns:
            A list of tuples containing:
            1. Integer discrete actions of shape ``[batch_size, 1]``.
            2. Action probability tensors of shape ``[batch_size, action_num]``.
            3. Any other things returned by your actor.
        """
        actions = self._act_api_general(states, use_target)
        result = []
        for action, *others in actions:
            assert_output_is_probs(action)
            batch_size = action.shape[0]
            action_disc = t.argmax(action, dim=1).view(batch_size, 1)
            result.append((action_disc, action, *others))
        return result

    def act_discrete_with_noise(
        self, states: List[Dict[str, Any]], use_target: bool = False
    ):
        """
        Use all actor networks to produce discrete actions for the current
        state.
        A random sub-policy from the policy ensemble of each actor will
        be chosen.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            states: A list of current states of each actor.
            use_target: Whether use the target network.

        Returns:
            A list of tuples containing:
            1. Integer noisy discrete actions.
            2. Action probability tensors of shape ``[batch_size, action_num]``.
            3. Any other things returned by your actor.
        """
        actions = self._act_api_general(states, use_target)
        result = []
        for action, *others in actions:
            assert_output_is_probs(action)
            batch_size = action.shape[0]
            dist = Categorical(action)
            action_disc = dist.sample([batch_size, 1]).view(batch_size, 1)
            result.append((action_disc, action, *others))
        return result

    def _act_api_general(self, states, use_target):
        if self.use_jit:
            if use_target:
                actors = [choice(sub_actors) for sub_actors in self.jit_actor_targets]
            else:
                actors = [choice(sub_actors) for sub_actors in self.jit_actors]
            future = [self._jit_safe_call(ac, st) for ac, st in zip(actors, states)]
            result = [t.jit._wait(fut) for fut in future]
            result = [res if isinstance(res, tuple) else (res,) for res in result]
        else:
            if use_target:
                actors = [choice(sub_actors) for sub_actors in self.actor_targets]
            else:
                actors = [choice(sub_actors) for sub_actors in self.actors]
            result = self.pool.starmap(self._no_grad_safe_call, zip(actors, states))
            result = [res for res in result]
        return result

    def _criticize(
        self,
        states: List[Dict[str, Any]],
        actions: List[Dict[str, Any]],
        index: int,
        use_target=False,
    ):
        """
        Use critic network to evaluate current value.

        Args:
            states: Current states of all actors.
            actions: Current actions of all actors.
            use_target: Whether to use the target network.
            index: Index of the used critic.

        Returns:
            Q Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(
                self.critic_targets[index],
                self.state_concat_function(states),
                self.action_concat_function(actions),
            )
        else:
            return safe_call(
                self.critics[index],
                self.state_concat_function(states),
                self.action_concat_function(actions),
            )

    def store_transitions(self, transitions: List[Union[Transition, Dict]]):
        """
        Add a list of transition samples, from all actors at the same time
        step, to the replay buffers.

        Args:
            transitions: List of transition objects.
        """
        assert len(transitions) == len(self.replay_buffers)
        for buff, trans in zip(self.replay_buffers, transitions):
            buff.append(
                trans,
                required_attrs=("state", "action", "next_state", "reward", "terminal"),
            )

    def store_episodes(self, episodes: List[List[Union[Transition, Dict]]]):
        """
        Add a List of full episodes, from all actors, to the replay buffers.
        Each episode is a list of transition samples.
        """
        assert len(episodes) == len(self.replay_buffers)
        all_length = [len(ep) for ep in episodes]
        assert len(set(all_length)) == 1, "All episodes must have the same length!"
        for buff, ep in zip(self.replay_buffers, episodes):
            for trans in ep:
                buff.append(
                    trans,
                    required_attrs=(
                        "state",
                        "action",
                        "next_state",
                        "reward",
                        "terminal",
                    ),
                )

    def update(
        self,
        update_value=True,
        update_policy=True,
        update_target=True,
        concatenate_samples=True,
    ):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.
            update_target: Whether to update targets.
            concatenate_samples: Whether to concatenate the samples.
        Returns:
            mean value of estimated policy value, value loss
        """
        # All buffers should have the same length now.

        # Create a sample method per update
        # this sample method will sample the same indexes
        # (different for each update() call) on all buffers.
        buffer_length = self.replay_buffers[0].size()
        if buffer_length == 0:
            return
        batch_size = min(buffer_length, self.batch_size)
        sample_indexes = [
            [randint(0, buffer_length - 1) for _ in range(batch_size)]
            for __ in range(self.ensemble_size)
        ]

        sample_methods = [
            self._create_sample_method(indexes) for indexes in sample_indexes
        ]

        # Now sample from buffer for each sub-policy in the ensemble.
        # To reduce memory usage, for each sub-policy "i" of each actor,
        # the same sample "i" will be used for training.

        # Tensors in the sampled batch will be moved to shared memory.

        # size: [ensemble size, num of actors]
        batches = []
        next_actions_t = []
        for e_idx in range(self.ensemble_size):
            ensemble_batch = []
            for a_idx in range(len(self.actors)):
                batch_size_, batch = self.replay_buffers[a_idx].sample_batch(
                    self.batch_size,
                    concatenate_samples,
                    sample_method=sample_methods[e_idx],
                    sample_attrs=[
                        "state",
                        "action",
                        "reward",
                        "next_state",
                        "terminal",
                        "*",
                    ],
                )
                ensemble_batch.append(batch)
                assert batch_size_ == batch_size

            batches.append(ensemble_batch)
            next_actions_t.append(
                [
                    self.action_transform_function(act)
                    for act in self.act(
                        [batch[3] for batch in ensemble_batch], target=True
                    )
                ]
            )

        if self.pool_type == "process":
            batches = self._move_to_shared_mem(batches)
            next_actions_t = self._move_to_shared_mem(next_actions_t)

        args = []
        self._update_counter += 1
        for e_idx in range(self.ensemble_size):
            for a_idx in range(len(self.actors)):
                args.append(
                    (
                        batch_size,
                        batches,
                        next_actions_t,
                        a_idx,
                        e_idx,
                        self.actors,
                        self.actor_targets,
                        self.critics,
                        self.critic_targets,
                        self.critic_visible_actors,
                        self.actor_optims,
                        self.critic_optims,
                        update_value,
                        update_policy,
                        update_target,
                        self.action_transform_function,
                        self.action_concat_function,
                        self.state_concat_function,
                        self.reward_function,
                        self.criterion,
                        self.discount,
                        self.update_rate,
                        self.update_steps,
                        self._update_counter,
                        self.grad_max,
                        self.visualize and not self.has_visualized,
                        self.visualize_dir,
                        self._backward,
                    )
                )
        all_loss = self.pool.starmap(self._update_sub_policy, args)
        mean_loss = t.tensor(all_loss).mean(dim=0)

        # returns action value and policy loss
        return -mean_loss[0].item(), mean_loss[1].item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_schs"):
            for actor_lr_sch in self.actor_lr_schs:
                actor_lr_sch.step()
        if hasattr(self, "critic_lr_schs"):
            for critic_lr_sch in self.critic_lr_schs:
                critic_lr_sch.step()

    def load(self, model_dir, network_map=None, version=-1):
        # DOC INHERITED
        super().load(model_dir, network_map, version)
        with t.no_grad():
            self.pool.starmap(
                hard_update,
                zip(
                    itertools.chain(*self.actors), itertools.chain(*self.actor_targets)
                ),
            )
            self.pool.starmap(hard_update, zip(self.critics, self.critic_targets))

    @staticmethod
    def _no_grad_safe_call(model, *named_args):
        with t.no_grad():
            result = safe_call(model, *named_args)
            return result

    @staticmethod
    def _jit_safe_call(model, *named_args):
        if not hasattr(model, "input_device") or not hasattr(model, "output_device"):
            # try to automatically determine the input & output
            # device of the model
            mt = type(model)
            device = determine_device(model)
            if len(device) > 1:
                raise RuntimeError(
                    f"""\
                    Failed to automatically determine i/o device of your model: {mt}
                    Detected multiple devices: {device}
    
                    You need to manually specify i/o device of your model.
    
                    Either Wrap your model of type nn.Module with one of:
                    1. static_module_wrapper from machin.model.nets.base
                    2. dynamic_module_wrapper from machin.model.nets.base 
                    
                    Or construct your own module & model with: 
                    NeuralNetworkModule from machin.model.nets.base"""
                )
            else:
                # assume that i/o devices are the same as parameter device
                # print a warning
                default_logger.warning(
                    f"""\
                    You have not specified the i/o device of your model {mt}
                    Automatically determined and set to: {device[0]}
    
                    The framework is not responsible for any un-matching device issues 
                    caused by this operation."""
                )
                model = static_module_wrapper(model, device[0], device[0])
        input_device = model.input_device
        # set in __init__
        args = model.arg_spec.args[1:] + model.arg_spec.kwonlyargs
        if model.arg_spec.defaults is not None:
            args_with_defaults = args[-len(model.arg_spec.defaults) :]
        else:
            args_with_defaults = []
        required_args = (
            set(args)
            - set(args_with_defaults)
            - set(
                model.arg_spec.kwonlydefaults.keys()
                if model.arg_spec.kwonlydefaults is not None
                else []
            )
        )
        model_type = model.model_type
        # t.jit._fork does not support keyword args
        # fill arguments in by their positions.
        args_list = [None for _ in args]
        args_filled = [False for _ in args]

        for na in named_args:
            for k, v in na.items():
                if k in args:
                    if k not in args:
                        pass
                    args_filled[args.index(k)] = True
                    if t.is_tensor(v):
                        args_list[args.index(k)] = v.to(input_device)
                    else:
                        args_list[args.index(k)] = v

        if not all(args_filled):
            not_filled = [arg for filled, arg in zip(args_filled, args) if not filled]
            req_not_filled = set(not_filled).intersection(required_args)
            if len(req_not_filled) > 0:
                raise RuntimeError(
                    f"""\
                    Required arguments of the forward function of Model {model_type} 
                    is {required_args}, missing required arguments: {req_not_filled}
        
                    Check your storage functions.
                    """
                )

        return t.jit._fork(model, *args_list)

    @staticmethod
    def _update_sub_policy(
        batch_size,
        batches,
        next_actions_t,
        actor_index,
        policy_index,
        actors,
        actor_targets,
        critics,
        critic_targets,
        critic_visible_actors,
        actor_optims,
        critic_optims,
        update_value,
        update_policy,
        update_target,
        atf,
        acf,
        scf,
        rf,
        criterion,
        discount,
        update_rate,
        update_steps,
        update_counter,
        grad_max,
        visualize,
        visualize_dir,
        backward_func,
    ):
        # atf: action transform function, used to transform the
        #      raw output of a single actor to a arg dict like:
        #      {"action": tensor}, where "action" is the keyword argument
        #      name of the critic.
        #
        # acf: action concatenation function, used to concatenate
        #      a list of action dicts into a single arg dict readable
        #      by critic.
        # scf: state concatenation function, used to concatenate
        #      a list of state dicts into a single arg dict readable
        #      by critic.
        # rf: reward function

        # The innermost element of ``batches``:
        # (state, action, reward, next_state, terminal, *)
        # ``batches`` size: [ensemble_size, actor_num]
        # select the batch for this sub-policy in the ensemble
        ensemble_batch = batches[policy_index]
        ensemble_n_act_t = next_actions_t[policy_index]
        visible_actors = critic_visible_actors[actor_index]

        actors[actor_index][policy_index].train()
        critics[actor_index].train()

        with t.no_grad():
            # only select visible actors
            all_next_actions_t = [
                ensemble_n_act_t[a_idx]
                if a_idx != actor_index
                else atf(
                    safe_call(
                        actor_targets[actor_index][policy_index],
                        ensemble_batch[a_idx][3],
                    )[0],
                    ensemble_batch[a_idx][5],
                )
                for a_idx in visible_actors
            ]
            all_next_actions_t = acf(all_next_actions_t)

            all_actions = [ensemble_batch[a_idx][1] for a_idx in visible_actors]
            all_actions = acf(all_actions)

            all_next_states = [ensemble_batch[a_idx][3] for a_idx in visible_actors]
            all_next_states = scf(all_next_states)

            all_states = [ensemble_batch[a_idx][0] for a_idx in visible_actors]
            all_states = scf(all_states)

        # Update critic network first
        # Generate target value using target critic.
        with t.no_grad():
            reward = ensemble_batch[actor_index][2]
            terminal = ensemble_batch[actor_index][4]
            next_value = safe_call(
                critic_targets[actor_index], all_next_states, all_next_actions_t
            )[0]
            next_value = next_value.view(batch_size, -1)
            y_i = rf(
                reward, discount, next_value, terminal, ensemble_batch[actor_index][5]
            )

        cur_value = safe_call(critics[actor_index], all_states, all_actions)[0]
        value_loss = criterion(cur_value, y_i.to(cur_value.device))

        if visualize:
            # only invoked if not running by pool
            MADDPG._visualize(value_loss, f"critic_{actor_index}", visualize_dir)

        if update_value:
            critics[actor_index].zero_grad()
            backward_func(value_loss)
            nn.utils.clip_grad_norm_(critics[actor_index].parameters(), grad_max)
            critic_optims[actor_index].step()

        # Update actor network
        all_actions = [ensemble_batch[a_idx][1] for a_idx in visible_actors]
        # find the actor index in the view range of critic
        # Eg: there are 4 actors in total: a_0, a_1, a_2, a_3
        # critic may have access to actor a_1 and a_2
        # then:
        #     visible_actors.index(a_1) = 0
        #     visible_actors.index(a_2) = 1
        # visible_actors.index returns the (critic-)local position of actor
        # in the view range of its corresponding critic.
        all_actions[visible_actors.index(actor_index)] = atf(
            safe_call(
                actors[actor_index][policy_index], ensemble_batch[actor_index][3]
            )[0],
            ensemble_batch[actor_index][5],
        )
        all_actions = acf(all_actions)

        act_value = safe_call(critics[actor_index], all_states, all_actions)[0]

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        if visualize:
            # only invoked if not running by pool
            MADDPG._visualize(
                act_policy_loss, f"actor_{actor_index}_{policy_index}", visualize_dir,
            )

        if update_policy:
            actors[actor_index][policy_index].zero_grad()
            backward_func(act_policy_loss)
            nn.utils.clip_grad_norm_(
                actors[actor_index][policy_index].parameters(), grad_max
            )
            actor_optims[actor_index][policy_index].step()

        # Update target networks
        if update_target:
            if update_rate is not None:
                soft_update(
                    actor_targets[actor_index][policy_index],
                    actors[actor_index][policy_index],
                    update_rate,
                )
                soft_update(
                    critic_targets[actor_index], critics[actor_index], update_rate
                )
            else:
                if update_counter % update_steps == 0:
                    hard_update(
                        actor_targets[actor_index][policy_index],
                        actors[actor_index][policy_index],
                    )
                    hard_update(critic_targets[actor_index], critics[actor_index])

        actors[actor_index][policy_index].eval()
        critics[actor_index].eval()
        return -act_policy_loss.item(), value_loss.item()

    @staticmethod
    def _visualize(final_tensor, name, directory):
        g = make_dot(final_tensor)
        g.render(
            filename=name, directory=directory, view=False, cleanup=False, quiet=True
        )

    @staticmethod
    def _move_to_shared_mem(obj):
        if t.is_tensor(obj):
            obj = obj.detach()
            obj.share_memory_()
            return obj
        elif isinstance(obj, list):
            for idx, sub_obj in enumerate(obj):
                obj[idx] = MADDPG._move_to_shared_mem(sub_obj)
            return obj
        elif isinstance(obj, tuple):
            obj = list(obj)
            for idx, sub_obj in enumerate(obj):
                obj[idx] = MADDPG._move_to_shared_mem(sub_obj)
            return tuple(obj)
        elif isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = MADDPG._move_to_shared_mem(v)
            return obj

    @staticmethod
    def _check_parameters_device(models):
        devices = set()
        for model in models:
            for k, v in model.named_parameters():
                devices.add(v.device.type)
                if len(devices) > 1:
                    raise RuntimeError(
                        "All of your models should either"
                        "locate on GPUs or on your CPU!"
                    )
        return list(devices)[0]

    @staticmethod
    def _create_sample_method(indexes):
        def sample_method(buffer, _len):
            nonlocal indexes
            batch = [buffer[i] for i in indexes if i < len(buffer)]
            return len(batch), batch

        return sample_method

    @staticmethod
    def action_transform_function(raw_output_action: Any, *_):
        return {"action": raw_output_action}

    @staticmethod
    def action_concat_function(actions: List[Dict], *_):
        # Assume an atom action is [batch_size, action_dim]
        # concatenate actions in the second dimension.
        # becomes [batch_size, actor_num * action_dim]
        keys = actions[0].keys()
        all_actions = {}
        for k in keys:
            all_actions[k] = t.cat([act[k].cpu() for act in actions], dim=1)
        return all_actions

    @staticmethod
    def state_concat_function(states: List[Dict], *_):
        # Assume an atom state is [batch_size, state_dim]
        # concatenate states in the second dimension.
        # becomes [batch_size, actor_num * state_dim]
        keys = states[0].keys()
        all_states = {}
        for k in keys:
            all_states[k] = t.cat([st[k].cpu() for st in states], dim=1)
        return all_states

    @staticmethod
    def reward_function(reward, discount, next_value, terminal, *_):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * ~terminal * next_value

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "models": [["Actor"], ["Actor"], ["Critic"], ["Critic"]],
            "model_args": ([()], [()], [()], [()]),
            "model_kwargs": ([{}], [{}], [{}], [{}]),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "critic_visible_actors": None,
            "sub_policy_num": 0,
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "update_rate": 0.001,
            "update_steps": None,
            "actor_learning_rate": 0.0005,
            "critic_learning_rate": 0.001,
            "discount": 0.99,
            "gradient_max": np.inf,
            "replay_size": 500000,
            "replay_device": "cpu",
            "replay_buffer": None,
            "visualize": False,
            "visualize_dir": "",
            "use_jit": True,
            "pool_type": "thread",
            "pool_size": None,
        }
        config = deepcopy(config)
        config["frame"] = "MADDPG"
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        f_config = deepcopy(config["frame_config"])
        all_models = []
        for models, model_args, model_kwargs in zip(
            f_config["models"], f_config["model_args"], f_config["model_kwargs"]
        ):
            models = assert_and_get_valid_models(models)
            models = [
                m(*arg, **kwarg).to(model_device)
                for m, arg, kwarg in zip(models, model_args, model_kwargs)
            ]
            all_models.append(models)
        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        f_config["optimizer"] = optimizer
        f_config["criterion"] = criterion
        f_config["lr_scheduler"] = lr_scheduler
        frame = cls(*all_models, **f_config)
        return frame



import warnings
from .buffer import Buffer
from .prioritized_buffer import WeightTree, PrioritizedBuffer

try:
    from .buffer_d import DistributedBuffer
    from .prioritized_buffer_d import DistributedPrioritizedBuffer
except ImportError as _:
    warnings.warn(
        "Failed to import buffers relying on torch.distributed." " Set them to None."
    )
    DistributedBuffer = None
    DistributedPrioritizedBuffer = None

__all__ = [
    "Buffer",
    "DistributedBuffer",
    "PrioritizedBuffer",
    "DistributedPrioritizedBuffer",
    "WeightTree",
]



from .dqn_per import *
from .ddpg_per import *
from ..buffers.prioritized_buffer_d import DistributedPrioritizedBuffer
from torch.nn.parallel import DistributedDataParallel
from machin.parallel.server import PushPullModelServer
from machin.parallel.distributed import get_world, RpcGroup
from machin.frame.helpers.servers import model_server_helper


def _disable_update(*_, **__):
    return None, None


class DQNApex(DQNPer):
    """
    Massively parallel version of a Double DQN with prioritized replay.

    The pull function is invoked before using ``act_discrete``,
    ``act_discrete_with_noise`` and ``criticize``.

    The push function is invoked after ``update``.
    """

    def __init__(
        self,
        qnet: Union[NeuralNetworkModule, nn.Module],
        qnet_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        apex_group: RpcGroup,
        model_server: Tuple[PushPullModelServer],
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple] = (),
        lr_scheduler_kwargs: Tuple[Dict] = (),
        batch_size: int = 100,
        epsilon_decay: float = 0.9999,
        update_rate: float = 0.005,
        update_steps: Union[int, None] = None,
        learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        **__,
    ):
        """
        See Also:
            :class:`.DQNPer`

        Note:
            Apex framework supports multiple workers(samplers), and only
            one trainer, you may use ``DistributedDataParallel`` in trainer.
            If you use ``DistributedDataParallel``, you must call ``update()``
            in all member processes of ``DistributedDataParallel``.

        Args:
            qnet: Q network module.
            qnet_target: Target Q network module.
            optimizer: Optimizer used to optimize ``qnet``.
            criterion: Criterion used to evaluate the value loss.
            apex_group: Group of all processes using the apex-DQN framework,
                including all samplers and trainers.
            model_server: Custom model sync server accessor for ``qnet``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            epsilon_decay: Epsilon decay rate per acting with noise step.
                ``epsilon`` attribute is multiplied with this every time
                ``act_discrete_with_noise`` is called.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            update_steps: Training step number used to update target networks.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            gradient_max: Maximum gradient.
            replay_size: Local replay buffer size of a single worker.
        """
        super().__init__(
            qnet,
            qnet_target,
            optimizer,
            criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            epsilon_decay=epsilon_decay,
            update_rate=update_rate,
            update_steps=update_steps,
            learning_rate=learning_rate,
            discount=discount,
            gradient_max=gradient_max,
        )
        self._is_using_DP_or_DDP = isinstance(
            self.qnet, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel)
        )

        # will not support sharing rpc group,
        # use static buffer_name is ok here.
        self.replay_buffer = DistributedPrioritizedBuffer(
            buffer_name="buffer", group=apex_group, buffer_size=replay_size
        )
        self.apex_group = apex_group
        self.qnet_model_server = model_server[0]
        self.is_syncing = True

    @classmethod
    def is_distributed(cls):
        return True

    def set_sync(self, is_syncing):
        self.is_syncing = is_syncing

    def manual_sync(self):
        if not self._is_using_DP_or_DDP:
            self.qnet_model_server.pull(self.qnet)

    def act_discrete(self, state: Dict[str, Any], use_target: bool = False, **__):
        # DOC INHERITED
        if self.is_syncing and not use_target and not self._is_using_DP_or_DDP:
            self.qnet_model_server.pull(self.qnet)
        return super().act_discrete(state, use_target)

    def act_discrete_with_noise(
        self,
        state: Dict[str, Any],
        use_target: bool = False,
        decay_epsilon: bool = True,
        **__,
    ):
        # DOC INHERITED
        if self.is_syncing and not use_target and not self._is_using_DP_or_DDP:
            self.qnet_model_server.pull(self.qnet)
        return super().act_discrete_with_noise(state, use_target, decay_epsilon)

    def update(
        self, update_value=True, update_target=True, concatenate_samples=True, **__
    ):
        # DOC INHERITED
        result = super().update(update_value, update_target, concatenate_samples)
        if self._is_using_DP_or_DDP:
            self.qnet_model_server.push(self.qnet.module, pull_on_fail=False)
        else:
            self.qnet_model_server.push(self.qnet)
        return result

    @classmethod
    def generate_config(cls, config: Dict[str, Any]):
        default_values = {
            "learner_process_number": 1,
            "model_server_group_name": "dqn_apex_model_server",
            "model_server_members": "all",
            "apex_group_name": "dqn_apex",
            "apex_members": "all",
            "models": ["QNet", "QNet"],
            "model_args": ((), ()),
            "model_kwargs": ({}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "epsilon_decay": 0.9999,
            "update_rate": 0.005,
            "update_steps": None,
            "learning_rate": 0.001,
            "discount": 0.99,
            "gradient_max": np.inf,
            "replay_size": 500000,
        }
        config = deepcopy(config)
        config["frame"] = "DQNApex"
        config["batch_num"] = {"sampler": 10, "learner": 1}
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        world = get_world()
        f_config = deepcopy(config["frame_config"])
        apex_group = world.create_rpc_group(
            group_name=f_config["apex_group_name"],
            members=(
                world.get_members()
                if f_config["apex_members"] == "all"
                else f_config["apex_members"]
            ),
        )

        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        # wrap models in DistributedDataParallel when running in learner mode
        max_learner_id = f_config["learner_process_number"]

        learner_group = world.create_collective_group(ranks=list(range(max_learner_id)))

        if world.rank < max_learner_id:
            models = [
                DistributedDataParallel(module=m, process_group=learner_group.group)
                for m in models
            ]

        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        criterion.reduction = "none"
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        servers = model_server_helper(
            model_num=1,
            group_name=f_config["model_server_group_name"],
            members=f_config["model_server_members"],
        )
        del f_config["optimizer"]
        del f_config["criterion"]
        del f_config["lr_scheduler"]
        frame = cls(
            *models,
            optimizer,
            criterion,
            apex_group,
            servers,
            lr_scheduler=lr_scheduler,
            **f_config,
        )
        if world.rank >= max_learner_id:
            frame.role = "sampler"
            frame.update = _disable_update
        else:
            frame.role = "learner"
        return frame


class DDPGApex(DDPGPer):
    """
    Massively parallel version of a DDPG with prioritized replay.

    The pull function is invoked before using
    ``act``, ``act_with_noise``, ``act_discrete``,
    ``act_discrete_with_noise`` and ``criticize``.

    The push function is invoked after ``update``.
    """

    def __init__(
        self,
        actor: Union[NeuralNetworkModule, nn.Module],
        actor_target: Union[NeuralNetworkModule, nn.Module],
        critic: Union[NeuralNetworkModule, nn.Module],
        critic_target: Union[NeuralNetworkModule, nn.Module],
        optimizer: Callable,
        criterion: Callable,
        apex_group: RpcGroup,
        model_server: Tuple[PushPullModelServer],
        *_,
        lr_scheduler: Callable = None,
        lr_scheduler_args: Tuple[Tuple, Tuple] = (),
        lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
        batch_size: int = 100,
        update_rate: float = 0.005,
        update_steps: Union[int, None] = None,
        actor_learning_rate: float = 0.0005,
        critic_learning_rate: float = 0.001,
        discount: float = 0.99,
        gradient_max: float = np.inf,
        replay_size: int = 500000,
        **__,
    ):
        """
        See Also:
            :class:`.DDPGPer`

        TODO:
            implement truncated n-step returns, just like the one used in
            :class:`.RAINBOW`.

        Note:
            Apex framework supports multiple workers(samplers), and only
            one trainer, you may use ``DistributedDataParallel`` in trainer.
            If you use ``DistributedDataParallel``, you must call ``update()``
            in all member processes of ``DistributedDataParallel``.

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            optimizer: Optimizer used to optimize ``qnet``.
            criterion: Criterion used to evaluate the value loss.
            apex_group: Group of all processes using the apex-DDPG framework,
                including all samplers and trainers.
            model_server: Custom model sync server accessor for ``actor``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            update_steps: Training step number used to update target networks.
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            gradient_max: Maximum gradient.
            replay_size: Local replay buffer size of a single worker.
        """
        super().__init__(
            actor,
            actor_target,
            critic,
            critic_target,
            optimizer,
            criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            update_rate=update_rate,
            update_steps=update_steps,
            actor_learning_rate=actor_learning_rate,
            critic_learning_rate=critic_learning_rate,
            discount=discount,
            gradient_max=gradient_max,
        )
        self._is_using_DP_or_DDP = isinstance(
            self.actor, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel)
        )
        # will not support sharing rpc group,
        # use static buffer_name is ok here.
        self.replay_buffer = DistributedPrioritizedBuffer(
            buffer_name="buffer", group=apex_group, buffer_size=replay_size
        )
        self.apex_group = apex_group
        self.actor_model_server = model_server[0]
        self.is_syncing = True

    @classmethod
    def is_distributed(cls):
        return True

    def set_sync(self, is_syncing):
        self.is_syncing = is_syncing

    def manual_sync(self):
        if not self._is_using_DP_or_DDP:
            self.actor_model_server.pull(self.actor)

    def act(self, state: Dict[str, Any], use_target: bool = False, **__):
        # DOC INHERITED
        if self.is_syncing and not use_target and not self._is_using_DP_or_DDP:
            self.actor_model_server.pull(self.actor)
        return super().act(state, use_target)

    def act_with_noise(
        self,
        state: Dict[str, Any],
        noise_param: Tuple = (0.0, 1.0),
        ratio: float = 1.0,
        mode: str = "uniform",
        use_target: bool = False,
        **__,
    ):
        # DOC INHERITED
        if self.is_syncing and not use_target and not self._is_using_DP_or_DDP:
            self.actor_model_server.pull(self.actor)
        return super().act_with_noise(
            state,
            noise_param=noise_param,
            ratio=ratio,
            mode=mode,
            use_target=use_target,
        )

    def act_discrete(self, state: Dict[str, Any], use_target: bool = False, **__):
        # DOC INHERITED
        if self.is_syncing and not use_target and not self._is_using_DP_or_DDP:
            self.actor_model_server.pull(self.actor)
        return super().act_discrete(state, use_target)

    def act_discrete_with_noise(
        self, state: Dict[str, Any], use_target: bool = False, **__
    ):
        # DOC INHERITED
        if self.is_syncing and not use_target and not self._is_using_DP_or_DDP:
            self.actor_model_server.pull(self.actor)
        return super().act_discrete_with_noise(state, use_target)

    def update(
        self,
        update_value=True,
        update_policy=True,
        update_target=True,
        concatenate_samples=True,
        **__,
    ):
        # DOC INHERITED
        result = super().update(
            update_value, update_policy, update_target, concatenate_samples
        )
        if self._is_using_DP_or_DDP:
            self.actor_model_server.push(self.actor.module, pull_on_fail=False)
        else:
            self.actor_model_server.push(self.actor)
        return result

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        default_values = {
            "learner_process_number": 1,
            "model_server_group_name": "ddpg_apex_model_server",
            "model_server_members": "all",
            "apex_group_name": "ddpg_apex",
            "apex_members": "all",
            "models": ["Actor", "Actor", "Critic", "Critic"],
            "model_args": ((), (), (), ()),
            "model_kwargs": ({}, {}, {}, {}),
            "optimizer": "Adam",
            "criterion": "MSELoss",
            "criterion_args": (),
            "criterion_kwargs": {},
            "lr_scheduler": None,
            "lr_scheduler_args": None,
            "lr_scheduler_kwargs": None,
            "batch_size": 100,
            "update_rate": 0.005,
            "update_steps": None,
            "learning_rate": 0.001,
            "discount": 0.99,
            "gradient_max": np.inf,
            "replay_size": 500000,
        }
        config = deepcopy(config)
        config["frame"] = "DDPGApex"
        config["batch_num"] = {"sampler": 10, "learner": 1}
        if "frame_config" not in config:
            config["frame_config"] = default_values
        else:
            config["frame_config"] = {**config["frame_config"], **default_values}
        return config

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        world = get_world()
        f_config = deepcopy(config["frame_config"])
        apex_group = world.create_rpc_group(
            group_name=f_config["apex_group_name"],
            members=(
                world.get_members()
                if f_config["apex_members"] == "all"
                else f_config["apex_members"]
            ),
        )

        models = assert_and_get_valid_models(f_config["models"])
        model_args = f_config["model_args"]
        model_kwargs = f_config["model_kwargs"]
        models = [
            m(*arg, **kwarg).to(model_device)
            for m, arg, kwarg in zip(models, model_args, model_kwargs)
        ]
        # wrap models in DistributedDataParallel when running in learner mode
        max_learner_id = f_config["learner_process_number"]

        learner_group = world.create_collective_group(ranks=list(range(max_learner_id)))

        if world.rank < max_learner_id:
            models = [
                DistributedDataParallel(module=m, process_group=learner_group.group)
                for m in models
            ]

        optimizer = assert_and_get_valid_optimizer(f_config["optimizer"])
        criterion = assert_and_get_valid_criterion(f_config["criterion"])(
            *f_config["criterion_args"], **f_config["criterion_kwargs"]
        )
        criterion.reduction = "none"
        lr_scheduler = f_config["lr_scheduler"] and assert_and_get_valid_lr_scheduler(
            f_config["lr_scheduler"]
        )
        servers = model_server_helper(
            model_num=1,
            group_name=f_config["model_server_group_name"],
            members=f_config["model_server_members"],
        )
        del f_config["optimizer"]
        del f_config["criterion"]
        del f_config["lr_scheduler"]
        frame = cls(
            *models,
            optimizer,
            criterion,
            apex_group,
            servers,
            lr_scheduler=lr_scheduler,
            **f_config,
        )
        if world.rank >= max_learner_id:
            frame.role = "sampler"
            frame.update = _disable_update
        else:
            frame.role = "learner"
        return frame



"""
Attributes:
    default_logger: The default global logger.

TODO: maybe add logging utilities for distributed scenario?
"""
import colorlog
from logging import INFO


class FakeLogger:
    def setLevel(self, level):
        pass

    def debug(self, msg, *args, **kwargs):
        pass

    def info(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def warn(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass

    def exception(self, msg, *args, exc_info=True, **kwargs):
        pass

    def critical(self, msg, *args, **kwargs):
        pass

    def log(self, level, msg, *args, **kwargs):
        pass


_default_handler = colorlog.StreamHandler()
_default_handler.setFormatter(
    colorlog.ColoredFormatter(
        "%(log_color)s[%(asctime)s] <%(levelname)s>:%(name)s:%(message)s"
    )
)

default_logger = colorlog.getLogger("default_logger")
default_logger.addHandler(_default_handler)
default_logger.setLevel(INFO)
fake_logger = FakeLogger()



import json
import argparse
from pprint import pprint
from machin.auto.config import (
    get_available_algorithms,
    get_available_environments,
    generate_algorithm_config,
    generate_env_config,
    generate_training_config,
    launch,
)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")

    p_list = subparsers.add_parser(
        "list", help="List available algorithms or environments."
    )

    p_list.add_argument(
        "--algo", action="store_true", help="List available algorithms.",
    )

    p_list.add_argument(
        "--env", action="store_true", help="List available environments."
    )

    p_generate = subparsers.add_parser("generate", help="Generate configuration.")

    p_generate.add_argument(
        "--algo", type=str, required=True, help="Algorithm name to use."
    )
    p_generate.add_argument(
        "--env", type=str, required=True, help="Environment name to use."
    )
    p_generate.add_argument(
        "--print", action="store_true", help="Direct config output to screen."
    )
    p_generate.add_argument(
        "--output",
        type=str,
        default="config.json",
        help="JSON config file output path.",
    )

    p_launch = subparsers.add_parser(
        "launch", help="Launch training with pytorch-lightning."
    )

    p_launch.add_argument(
        "--config", type=str, default="config.json", help="JSON config file path.",
    )

    args = parser.parse_args()
    if args.command == "list":
        if args.env:
            print("Available environments are:")
            for env in get_available_environments():
                print(env)
        elif args.algo:
            print("Available algorithms are:")
            for algo in get_available_algorithms():
                print(algo)
        else:
            print("You can list --algo or --env.")

    elif args.command == "generate":
        if args.algo not in get_available_algorithms():
            print(
                f"{args.algo} is not a valid algorithm, use list "
                "--algo to get a list of available algorithms."
            )
            exit(0)
        if args.env not in get_available_environments():
            print(
                f"{args.env} is not a valid environment, use list "
                "--env to get a list of available environments."
            )
            exit(0)
        config = {}
        config = generate_env_config(args.env, config=config)
        config = generate_algorithm_config(args.algo, config=config)
        config = generate_training_config(config=config)

        if args.print:
            pprint(config)

        with open(args.output, "w") as f:
            json.dump(config, f, indent=4, sort_keys=True)
        print(f"Config saved to {args.output}")

    elif args.command == "launch":
        with open(args.config, "r") as f:
            conf = json.load(f)
        launch(conf)



import os
import shutil
from PIL.Image import Image
from matplotlib.figure import Figure
from argparse import Namespace
from typing import Union, Dict, Optional, Any
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.loggers import LightningLoggerBase
from pytorch_lightning.loggers.base import rank_zero_experiment


class LocalMediaLogger(LightningLoggerBase):
    """
    A logger designed to handle log_image and log_artifact just like the
    neptune.ai logger does. log_hyperparams and log_metrics are not handled.
    """

    def __init__(self, image_dir: str, artifact_dir: str):
        super().__init__()
        self.image_dir = image_dir
        self.artifact_dir = artifact_dir
        self._counters = {}

    @property
    def name(self):
        return "LocalMediaLogger"

    @property
    @rank_zero_experiment
    def experiment(self):
        return None

    @property
    def version(self):
        return "0.1"

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]):
        """
        Not implemented.
        """
        pass

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int]):
        """
        Not implemented.
        """
        pass

    @rank_zero_only
    def log_artifact(self, artifact: str, destination: Optional[str] = None):
        """
        Save an artifact (file) in the ``artifact_dir``.

        Args:
            artifact: A path to the file in local filesystem.
            destination: Optional. Default is ``None``. A destination path
                in the artifact directory.
                If ``None`` is passed, an artifact file name will be used.
        """
        if destination:
            shutil.copy(
                artifact,
                self.enumerate_till_valid(os.path.join(self.artifact_dir, destination)),
            )
        else:
            shutil.copy(
                artifact,
                self.enumerate_till_valid(
                    os.path.join(self.artifact_dir, os.path.basename(artifact))
                ),
            )

    @rank_zero_only
    def log_image(
        self, log_name: str, image: Union[str, Any], step: Optional[int] = None
    ) -> None:
        """
        Log image data in the ``image_dir``.

        Args:
            log_name: The name of log, i.e. bboxes, visualisations,
                sample_images.
            image: The value of the log (data-point).
                Can be one of the following types: PIL image,
                `matplotlib.figure.Figure`,
                path to image file (str).
            step: Step number at which the metrics should be recorded,
                must be strictly increasing.
        """
        if log_name not in self._counters:
            self._counters[log_name] = 0

        if not isinstance(image, str):
            log_path = log_name + f"_{step or self._counters[log_name]}.png"
        else:
            extension = os.path.splitext(image)[1]
            log_path = log_name + f"_{step or self._counters[log_name]}{extension}"
        self._counters[log_name] += 1

        path = os.path.join(self.image_dir, log_path)
        if isinstance(image, Image):
            image.save(path)
        elif isinstance(image, Figure):
            image.savefig(path)
        elif isinstance(image, str):
            shutil.copy(image, path)
        else:
            raise ValueError(f"Unsupported image type: {type(image)}")

    @rank_zero_only
    def save(self):
        super().save()

    @rank_zero_only
    def finalize(self, status):
        pass

    @staticmethod
    def enumerate_till_valid(path):
        counter = 0
        cur_path = path
        while os.path.exists(cur_path):
            counter += 1
            names = list(os.path.splitext(path))
            names[0] += f"_{counter}"
            cur_path = "".join(names)
        return cur_path



from typing import Callable, Any, Union, List, Tuple, Dict
from machin.parallel.distributed import get_world, get_cur_name
from machin.parallel.server import PushPullGradServerImpl, PushPullModelServerImpl
from torch.optim import Adam


def grad_server_helper(
    model_creators: List[Callable],
    group_name: str = "grad_server",
    members: Union[str, List[str]] = "all",
    optimizer: Any = Adam,
    learning_rate: Union[float, List[float]] = 1e-3,
    optimizer_kwargs: List[Dict[str, Any]] = None,
    lr_scheduler: Any = None,
    lr_scheduler_args: List[Tuple] = None,
    lr_scheduler_kwargs: List[Dict[str, Any]] = None,
):
    """
    Helper function for creating a tuple of grad servers,
    used by A3C, IMPALE, etc. This function requires all processes
    in the world to enter.

    Args:
        model_creators: A list of model creator functions,
            each one corresponds to one gradient reduction server.
        group_name: Name of the RPC group where gradient servers should be
            registered on, the group name should be unique.
        members: Name of the involved RPC processes, ``"all"`` for all
            processes, they will be used as secondary reducers, the first
            process will be the primary reducer.
        optimizer: Optimizer class, default is Adam.
        learning_rate: Learning rate of each optimizer. Or a single float value
            for every one.
        optimizer_kwargs: Optimizer keyword arguments for each optimizer of
            each model.
        lr_scheduler: Learning rate scheduler class.
        lr_scheduler_args: Learning rate scheduler arguments for each
            lr_scheduler corresponding to each optimizer.
        lr_scheduler_kwargs: Learning rate scheduler keyword arguments for each
            lr_scheduler corresponding to each optimizer.

    Returns:
        A tuple of accessors to gradient servers, the tuple has the
        same size as ``model_creators``
    """
    # Note:
    # passing a list of creator functions instead of passing a list of models
    # directly is designed to remove the unnecessary model creation cost on
    # not-the-primary-reducer processes.

    # create groups first
    world = get_world()
    members = world.get_members() if members == "all" else members
    server_group = world.create_rpc_group(group_name, members)

    if isinstance(learning_rate, float):
        learning_rate = [learning_rate] * len(model_creators)

    optimizer_kwargs = optimizer_kwargs or [{}] * len(model_creators)
    lr_scheduler_args = lr_scheduler_args or [()] * len(model_creators)
    lr_scheduler_kwargs = lr_scheduler_kwargs or [{}] * len(model_creators)

    # create servers
    primary_reducer = members[0]
    servers = [
        PushPullGradServerImpl(
            "grad_server_" + str(i), server_group, primary_reducer=primary_reducer
        )
        for i in range(len(model_creators))
    ]
    if get_cur_name() == primary_reducer:
        for (
            model_creator,
            server,
            optim_kwargs,
            lr,
            lr_sch_args,
            lr_sch_kwargs,
        ) in zip(
            model_creators,
            servers,
            optimizer_kwargs,
            learning_rate,
            lr_scheduler_args,
            lr_scheduler_kwargs,
        ):
            model = model_creator()
            if lr_scheduler is None:
                server.manage_model(
                    model, optimizer(model.parameters(), lr=lr, **optim_kwargs)
                )
            else:
                optimizer = optimizer(model.parameters(), lr=lr, **optim_kwargs)
                server.manage_model(
                    model,
                    optimizer,
                    lr_scheduler(optimizer, *lr_sch_args, **lr_sch_kwargs),
                )
            server.start()

    server_group.barrier()
    servers = tuple(
        server_group.get_paired("grad_server_" + str(i)).to_here()
        for i in range(len(model_creators))
    )

    # accessors instead of actual implementation instance
    # will be returned because of __reduce__
    return servers


def model_server_helper(
    model_num: int,
    group_name: str = "model_server",
    members: Union[str, List[str]] = "all",
):
    """
    Helper function for creating a tuple of model servers,
    used by APEX, etc. This function requires all processes
    in the world to enter.

    Args:
        model_num: The number of models, corresponds to the number of model
            servers, since each server manages 1 model.
        group_name: Name of the RPC group where gradient servers should be
            registered on, the group name should be unique.
        members: Name of the involved RPC processes, ``"all"`` for all
            processes, only the first process will serve as the server in the
            current implementation.

    Returns:
        A tuple of accessors to model servers, the size of tuple is
        ``model_num``
    """
    # create groups first
    world = get_world()
    members = world.get_members() if members == "all" else members
    server_group = world.create_rpc_group(group_name, members)

    # create servers
    # In current implementation, only one process will initialize the server
    if get_cur_name() == members[0]:
        for i in range(model_num):
            _server = PushPullModelServerImpl("model_server_" + str(i), server_group)

    server_group.barrier()

    servers = tuple(
        server_group.get_paired("model_server_" + str(i)).to_here()
        for i in range(model_num)
    )

    # accessors instead of actual implementation instance
    # will be returned because of __reduce__
    return servers



from typing import Union, Dict, List, Tuple, Any, Callable
from ..transition import (
    Transition,
    Scalar,
    TransitionStorageSmart,
    TransitionStorageBasic,
)
import torch as t
import random


class Buffer:
    def __init__(self, buffer_size, buffer_device="cpu", *_, **__):
        """
        Create a buffer instance.

        Buffer stores a series of transition objects and functions
        as a ring buffer. **It is not thread-safe**.

        See Also:
            :class:`.Transition`


        During sampling, the tensors in "state", "action" and "next_state"
        dictionaries, along with "reward", will be concatenated in dimension 0.
        any other custom keys specified in ``**kwargs`` will not be
        concatenated.

        Args:
            buffer_size: Maximum buffer size.
            buffer_device: Device where buffer is stored.
        """
        self.buffer_size = buffer_size
        self.buffer_device = buffer_device
        self.buffer = TransitionStorageSmart(buffer_size)
        self.index = 0

    def append(
        self,
        transition: Union[Transition, Dict],
        required_attrs=("state", "action", "next_state", "reward", "terminal"),
    ):
        """
        Store a transition object to buffer.

        Args:
            transition: A transition object.
            required_attrs: Required attributes. Could be an empty tuple if
                no attribute is required.

        Raises:
            ``ValueError`` if transition object doesn't have required
            attributes in ``required_attrs`` or has different attributes
            compared to other transition objects stored in buffer.
        """
        if isinstance(transition, dict):
            transition = Transition(**transition)
        elif isinstance(transition, Transition):
            pass
        else:  # pragma: no cover
            raise RuntimeError(
                "Transition object must be a dict or an instance"
                " of the Transition class"
            )
        if not transition.has_keys(required_attrs):
            missing_keys = set(required_attrs) - set(transition.keys())
            raise ValueError(f"Transition object missing attributes: {missing_keys}")
        transition.to(self.buffer_device)

        if self.size() != 0 and self.buffer[0].keys() != transition.keys():
            raise ValueError("Transition object has different attributes!")

        return self.buffer.store(transition)

    def size(self):
        """
        Returns:
            Length of current buffer.
        """
        return len(self.buffer)

    def clear(self):
        """
        Remove all entries from the buffer
        """
        self.buffer.clear()

    @staticmethod
    def sample_method_random_unique(
        buffer: List[Transition], batch_size: int
    ) -> Tuple[int, List[Transition]]:
        """
        Sample unique random samples from buffer.

        Note:
            Sampled size could be any value from 0 to ``batch_size``.
        """
        if len(buffer) < batch_size:
            batch = random.sample(buffer, len(buffer))
            real_num = len(buffer)
        else:
            batch = random.sample(buffer, batch_size)
            real_num = batch_size
        return real_num, batch

    @staticmethod
    def sample_method_random(
        buffer: List[Transition], batch_size: int
    ) -> Tuple[int, List[Transition]]:
        """
        Sample random samples from buffer.

        Note:
            Sampled size could be any value from 0 to ``batch_size``.
        """
        indexes = [random.randint(0, len(buffer) - 1) for _ in range(batch_size)]
        batch = [buffer[i] for i in indexes]
        return batch_size, batch

    @staticmethod
    def sample_method_all(buffer: List[Transition], _) -> Tuple[int, List[Transition]]:
        """
        Sample all samples from buffer. Always return the whole buffer,
        will ignore the ``batch_size`` parameter.
        """
        return len(buffer), buffer

    def sample_batch(
        self,
        batch_size: int,
        concatenate: bool = True,
        device: Union[str, t.device] = None,
        sample_method: Union[Callable, str] = "random_unique",
        sample_attrs: List[str] = None,
        additional_concat_attrs: List[str] = None,
        *_,
        **__,
    ) -> Any:
        """
        Sample a random batch from buffer.

        See Also:
            Default sample methods are defined as static class methods.

            :meth:`.Buffer.sample_method_random_unique`

            :meth:`.Buffer.sample_method_random`

            :meth:`.Buffer.sample_method_all`

        Note:
            "Concatenation"
            means ``torch.cat([...], dim=0)`` for tensors,
            and ``torch.tensor([...]).view(batch_size, 1)`` for scalars.

        Warnings:
            Custom attributes must not contain tensors. And only scalar custom
            attributes can be concatenated, such as ``int``, ``float``,
            ``bool``.

        Args:
            batch_size: A hint size of the result sample. actual sample size
                        depends on your sample method.
            sample_method: Sample method, could be one of:
                           ``"random", "random_unique", "all"``,
                           or a function:
                           ``func(list, batch_size)->(list, result_size)``
            concatenate: Whether concatenate state, action and next_state
                         in dimension 0.
                         If ``True``, for each value in dictionaries of major
                         attributes. and each value of sub attributes, returns
                         a concatenated tensor. Custom Attributes specified in
                         ``additional_concat_attrs`` will also be concatenated.
                         If ``False``, return a list of tensors.
            device:      Device to copy to.
            sample_attrs: If sample_keys is specified, then only specified keys
                         of the transition object will be sampled. You may use
                         ``"*"`` as a wildcard to collect remaining
                         **custom keys** as a ``dict``, you cannot collect major
                         and sub attributes using this.
                         Invalid sample attributes will be ignored.
            additional_concat_attrs: additional **custom keys** needed to be
                         concatenated, will only work if ``concatenate`` is
                         ``True``.

        Returns:
            1. Batch size, Sampled attribute values in the same order as
               ``sample_keys``.

            2. Sampled attribute values is a tuple. Or ``None`` if sampled
               batch size is zero (E.g.: if buffer is empty or your sample
               size is 0 and you are not sampling using the "all" method).

               - For major attributes, result are dictionaries of tensors with
                 the same keys in your transition objects.

               - For sub attributes, result are tensors.

               - For custom attributes, if they are not in
                 ``additional_concat_attrs``, then lists, otherwise tensors.
        """
        if isinstance(sample_method, str):
            if not hasattr(self, "sample_method_" + sample_method):
                raise RuntimeError(
                    f"Cannot find specified sample method: {sample_method}"
                )
            sample_method = getattr(self, "sample_method_" + sample_method)
        batch_size, batch = sample_method(self.buffer, batch_size)

        if device is None:
            device = self.buffer_device

        return (
            batch_size,
            self.post_process_batch(
                batch, device, concatenate, sample_attrs, additional_concat_attrs
            ),
        )

    @classmethod
    def post_process_batch(
        cls,
        batch: List[Transition],
        device: Union[str, t.device],
        concatenate: bool,
        sample_attrs: List[str],
        additional_concat_attrs: List[str],
    ):
        """
        Post-process (concatenate) sampled batch.
        """
        result = []
        used_keys = []

        if len(batch) == 0:
            return None
        if sample_attrs is None:
            sample_attrs = batch[0].keys() if batch else []
        if additional_concat_attrs is None:
            additional_concat_attrs = []

        major_attr = set(batch[0].major_attr)
        sub_attr = set(batch[0].sub_attr)
        custom_attr = set(batch[0].custom_attr)
        for attr in sample_attrs:
            if attr in major_attr:
                tmp_dict = {}
                for sub_k in batch[0][attr].keys():
                    tmp_dict[sub_k] = cls.make_tensor_from_batch(
                        [item[attr][sub_k].to(device) for item in batch],
                        device,
                        concatenate,
                    )
                result.append(tmp_dict)
                used_keys.append(attr)
            elif attr in sub_attr:
                result.append(
                    cls.make_tensor_from_batch(
                        [item[attr] for item in batch], device, concatenate
                    )
                )
                used_keys.append(attr)
            elif attr == "*":
                # select custom keys
                tmp_dict = {}
                for remain_k in batch[0].keys():
                    if (
                        remain_k not in major_attr
                        and remain_k not in sub_attr
                        and remain_k not in used_keys
                    ):
                        tmp_dict[remain_k] = cls.make_tensor_from_batch(
                            [item[remain_k] for item in batch],
                            device,
                            concatenate and remain_k in additional_concat_attrs,
                        )
                result.append(tmp_dict)
            elif attr in custom_attr:
                result.append(
                    cls.make_tensor_from_batch(
                        [item[attr] for item in batch],
                        device,
                        concatenate and attr in additional_concat_attrs,
                    )
                )
                used_keys.append(attr)
        return tuple(result)

    @staticmethod
    def make_tensor_from_batch(
        batch: List[Union[Scalar, t.Tensor]],
        device: Union[str, t.device],
        concatenate: bool,
    ):
        """
        Make a tensor from a batch of data.
        Will concatenate input tensors in dimension 0.
        Or create a tensor of size (batch_size, 1) for scalars.

        Args:
            batch: Batch data.
            device: Device to move data to
            concatenate: Whether performing concatenation.

        Returns:
            Original batch if batch is empty,
            or tensor depends on your data (if concatenate),
            or original batch (if not concatenate).
        """
        if concatenate and len(batch) != 0:
            item = batch[0]
            batch_size = len(batch)
            if t.is_tensor(item):
                batch = [it.to(device) for it in batch]
                return t.cat(batch, dim=0).to(device)
            else:
                try:
                    return t.tensor(batch, device=device).view(batch_size, -1)
                except Exception:
                    raise ValueError(f"Batch not concatenable: {batch}")
        else:
            return batch

    def __reduce__(self):
        # for pickling
        return self.__class__, (self.buffer_size, self.buffer_device)



from typing import Any, Union, List
from random import choice
from copy import deepcopy
from queue import Queue
from threading import Event

import enum
import torch as t
import torch.nn as nn

from machin.parallel.thread import Thread
from machin.parallel.distributed import RpcGroup
from machin.utils.prepare import prep_load_state_dict
from .ordered_server import (
    OrderedServerBase,
    OrderedServerSimple,
    OrderedServerSimpleImpl,
)


class PushPullModelServer:
    def __init__(self, model_name: str, o_server: OrderedServerBase = None):
        """
        Create an accessor to the services provided by
        :class:`PushPullModelServerImpl`

        Args:
            model_name: Name of the managed model in the ordered server,
                only needed if ``server`` needs such a identifier. The default
                ordered server does not require this.
            o_server: Ordered server accessor.
        """
        self.model_name = model_name
        self.o_server = o_server

    def push(self, model: nn.Module, pull_on_fail=True):
        """
        Try to push a model to the ordered server, if failed, the newest
        model will be automatically pulled and its parameters will be
        assigned to ``model``. Gradients will not be cleared.

        Args:
            model: Model to push.
            pull_on_fail: Pull the newest parameters if push failed.

        Returns:
            True if push succeeded, else False.
        """
        if not hasattr(model, "pp_version"):
            model.pp_version = 0

        copied_model_params = deepcopy(model.state_dict())
        for k, v in copied_model_params.items():
            copied_model_params[k] = v.to("cpu")
        if not self.o_server.push(
            self.model_name,
            copied_model_params,
            version=model.pp_version + 1,
            prev_version=model.pp_version,
        ):
            if pull_on_fail:
                result = self.o_server.pull(self.model_name)
                if result is None:  # pragma: no cover
                    raise RuntimeError("Pull failed, this should not happen.")
                st_dict, version = result
                prep_load_state_dict(model, st_dict)
                model.pp_version = version
            return False
        else:
            model.pp_version += 1
        return True

    def pull(self, model: nn.Module):
        """
        Pull the newest state dict of your model and update its parameters
        and ``pp_version``. Gradients will not be cleared.

        Args:
            model: Model to pull.

        Returns:
            True if pull succeeded, else False.
        """
        result = self.o_server.pull(self.model_name)
        if result is None:  # pragma: no cover
            return False
        st_dict, version = result
        if not hasattr(model, "pp_version") or model.pp_version < version:
            prep_load_state_dict(model, st_dict)
            model.pp_version = version
        return True


class PushPullModelServerImpl:
    """
    A simple parameter server, which synchronize model parameters
    by pushing and pulling all parameters and maintaining a strict
    ordered version chain.

    Warning:
        Only one model is supported.
    """

    def __init__(
        self,
        server_name: str,
        group: RpcGroup,
        model_name: str = "model",
        o_server: OrderedServerBase = None,
    ):
        """
        This init function must be only invoked on the runner process,
        and the runner process must be a member process of ``group``.

        Args:
            server_name: Name of this server, used to registered
                the server as a paired class of ``group``.
            group: RpcGroup of the default server :class:`.OrderedServerSimple`
                mutually exclusive with ``o_server``
            model_name: Name of the managed model in the ordered server,
                only needed if ``server`` needs such a identifier. The default
                ordered server does not require this.
            o_server: Custom ordered server accessor.
        """
        self.server_name = server_name
        self.group = group
        self.model_name = model_name
        # actual running server started by OrderedServerSimpleStarter
        self._o_server_impl = None
        if o_server is None:
            self._o_server_impl = OrderedServerSimpleImpl(
                server_name + "_o_server", group
            )
            self.o_server = group.get_paired(server_name + "_o_server").to_here()
        else:  # pragma: no cover
            self.o_server = o_server
        # pair an accessor to group
        self.group.pair(
            server_name, PushPullModelServer(self.model_name, self.o_server)
        )


class ReduceType(enum.Enum):
    REDUCE_PRIMARY = 0
    REDUCE_SECONDARY = 1


class PushPullGradServer:
    def __init__(
        self,
        server_name: str,
        group: RpcGroup,
        model_name: str,
        secondary_reducers: List[str],
        o_server: OrderedServerBase,
    ):
        self.group = group
        self.model_name = model_name
        self.o_server = o_server
        self.secondary_services = [
            server_name + "/" + m + "/_push_service" for m in secondary_reducers
        ]

    def push(self, model: nn.Module):
        """
        Push the gradients of your model, then pull the newest parameters.
         Its gradients will be cleared.

        Args:
            model: Model to push.

        Returns:
            True if push succeeded, else False.
        """
        # extract gradients from the model
        grad_dict = {}
        for k, v in model.named_parameters():
            if not hasattr(v, "grad") or not t.is_tensor(v.grad):  # pragma: no cover
                raise RuntimeError(f"Parameter {k} doesn't have gradient to push!")
            grad_dict[k] = deepcopy(v.grad).to("cpu")
        self.group.registered_sync(
            choice(self.secondary_services),
            args=(grad_dict, ReduceType.REDUCE_SECONDARY),
        )
        return self.pull(model)

    def pull(self, model: nn.Module):
        """
        Pull the newest model. Its gradients will be cleared.

        Args:
            model: Model to push.

        Returns:
            True if pull succeeded, else False.
        """
        model.zero_grad()
        params = self.o_server.pull(self.model_name)
        if params is not None:
            # params could be None if the master reducer has't performed
            # a single reduction operation yet
            prep_load_state_dict(model, params[0])
            return True
        else:
            return False


class PushPullGradServerImpl:
    """
    A simple parameter server, which synchronize model parameters
    by pushing gradients and pulling back new parameters, no strict
    order is guaranteed.

    Warning:
        ``DistributedDataParallel`` is not supported. since we cannot
        load state dictionary after creation.
    """

    REDUCE_MASTER = 0
    REDUCE_SLAVE = 1

    def __init__(
        self,
        server_name: str,
        group: RpcGroup,
        model_name: str = "model",
        primary_reducer: str = None,
        secondary_reducers: List[str] = None,
        o_server: OrderedServerBase = None,
        reduce_method: str = "sum",
        reduce_device: Union[t.device, str] = "cpu",
        reduce_batch_size: int = 4,
        max_queue_size: int = 64,
    ):
        """
        Note:
            You should initialize ``PushPullGradServer`` on all members of
            ``secondary_reducers``, and ``primary_reducer``. Both of them
            should be members of the ``group``.

        Note:
            Internally the primary reducer will push updated versions
            to the ordered server.

        Hint:
            Reduction is performed in a tree fashion:

            1. In the first step, clients will push new gradients to a
               random secondary reducer, and the secondary reducer will perform
               the first reduction pass, then secondary reducers will push
               their results to the primary reducer.
            2. In the second step, the primary reducer will reduce results
               from the secondary reducer to get the final reduced gradient
               dictionary (has the same structure as state_dict), and assign
               gradients to its **managed model**, and perform the
               optimization.
            3. In the final step, the primary reducer will push the final
               model to the model server group, then clients can pull the
               newest model.

        Args:
            server_name: Name of this server, used to registered
                the server as a paired class of ``group``.
            group: Server group.
            model_name: Name of the managed model in the ordered server,
                only needed if ``server`` needs such a identifier. The default
                ordered server does not require this.
            primary_reducer: Name of the process serving as the primary reducer,
                which collects reduced gradients from secondary reducers and
                perform the final reduction.
            secondary_reducers: Name of the process serving as secondary
                reducers.
            o_server: Custom ordered server accessor. By default, the ordered
                server is a :class:`.OrderedServerSimple` hosted on the primary
                reducer.
            reduce_method: "mean" or "sum"
            reduce_device: Device to perform reduction, by default it is "cpu".
            reduce_batch_size: Size of a single reduction batch, server will
                wait until the number of requests in the reduction queue have
                reached this size.
            max_queue_size: Maximum reduction request queue size.
        """
        self.server_name = server_name
        self.group = group
        self.model_name = model_name

        if primary_reducer is None:
            primary_reducer = group.get_group_members()[0]
        assert group.is_member(primary_reducer)
        assert group.is_member()

        # actual running server started by OrderedServerSimpleStarter
        self._o_server_impl = None
        self.o_server = None
        if o_server is None:
            if group.get_cur_name() == primary_reducer:
                self._o_server_impl = OrderedServerSimpleImpl(
                    server_name + "_o_server", group
                )
            self.o_server = OrderedServerSimple(server_name + "_o_server", group)
        else:  # pragma: no cover
            self.o_server = o_server

        if secondary_reducers is None:
            secondary_reducers = group.get_group_members()

        self.primary_reducer = primary_reducer
        self.primary_service = server_name + "/" + primary_reducer + "/_push_service"
        self.secondary_reducers = secondary_reducers
        self.secondary_services = [
            server_name + "/" + m + "/_push_service" for m in secondary_reducers
        ]
        # register secondary reducer service
        self.group.register(
            server_name + "/" + group.get_cur_name() + "/_push_service",
            self._push_service,
        )

        # pair an accessor to group
        if self.group.get_cur_name() == self.primary_reducer:
            self.group.pair(
                self.server_name,
                PushPullGradServer(
                    self.server_name,
                    self.group,
                    self.model_name,
                    self.secondary_reducers,
                    self.o_server,
                ),
            )

        # prepare to start the reduction sub-thread
        assert reduce_method in ("mean", "sum")
        assert max_queue_size > 1
        assert reduce_batch_size > 1
        assert max_queue_size > reduce_batch_size
        self.started = False
        self.reduce_method = reduce_method
        self.reduce_batch_size = reduce_batch_size
        self.reduce_device = reduce_device
        self.max_queue_size = max_queue_size
        self.model = None  # type: Union[nn.Module, None]
        self.optimizer = None
        self.lr_scheduler = None
        # do not set max_queue_size here, will raise queue.Full
        self.master_queue = Queue()
        self.secondary_queue = Queue()
        self.work_event = Event()
        self.stop_event = Event()
        self.reduce_task = Thread(target=self._task_reduce_grad)
        self.reduce_task.daemon = True

    def start(self):
        if not self.started:
            self.reduce_task.start()
            self.started = True

    def stop(self):
        if self.started:
            self.stop_event.set()
            self.reduce_task.join()
            self.stop_event.clear()

    def watch(self):
        self.reduce_task.watch()

    def manage_model(self, model: nn.Module, optimizer: Any, lr_scheduler: Any = None):
        """
        Let the main reducer manage your model. Must be called before start.

        Warning:
            Make sure that the managed model is different from the model
            you use in your algorithms such as A3C!

        Args:
            model: Model to manage.
            optimizer: Optimizer of your model. you should initialize it first:
            >>> optimizer = Adam(model.parameters(), lr=1e-3)

            lr_scheduler: learning rate scheduler, you should initialize it
                first:
            >>> scheduler = LambdaLR(optimizer, lr_lambda=[lambda1, lambda2])

        Raises:
            ``RuntimeError`` if current rpc role is not the main reducer.
        """
        if self.group.get_cur_name() == self.primary_reducer:
            self.model = model
            self.optimizer = optimizer
            self.lr_scheduler = lr_scheduler
            self.model.pp_version = 0
        else:  # pragma: no cover
            raise RuntimeError(
                "Current worker is not the reduce master, and"
                "cannot manage the model."
            )

    def _push_service(self, grad_dict, level):  # pragma: no cover
        # Append reduce requests to queue.
        if level == ReduceType.REDUCE_SECONDARY:
            self.secondary_queue.put_nowait(grad_dict)
            self.work_event.set()
            self.work_event.clear()
        elif level == ReduceType.REDUCE_PRIMARY:
            self.master_queue.put_nowait(grad_dict)
            self.work_event.set()
            self.work_event.clear()
        else:  # pragma: no cover
            raise ValueError(f"Unknown push level: {level}")

    def _task_reduce_grad(self):
        while True:
            # Wait until one queue has reached target batch size
            while (
                self.master_queue.qsize() < self.reduce_batch_size
                and self.secondary_queue.qsize() < self.reduce_batch_size
            ):
                self.work_event.wait(timeout=1e-1)
                if self.stop_event.is_set():
                    return
            # discard oldest messages
            while self.master_queue.qsize() > self.max_queue_size:
                self.master_queue.get()
            while self.secondary_queue.qsize() > self.max_queue_size:
                self.secondary_queue.get()

            if self.master_queue.qsize() >= self.reduce_batch_size:
                # Perform reduction on the master reduction queue
                # Only the master reducer will execute this branch
                grad_dict = self._reduce_batch(
                    self.master_queue,
                    self.reduce_batch_size,
                    self.reduce_method,
                    self.reduce_device,
                )
                # Assign gradients to the managed model and
                # perform optimization.
                if self.model is not None and self.optimizer is not None:
                    self.optimizer.zero_grad()
                    with t.no_grad():
                        for k, v in self.model.named_parameters():
                            v.grad = grad_dict[k].to(v.device)
                    self.optimizer.step()
                    self.o_server.push(
                        self.model_name,
                        self.model.to("cpu").state_dict(),
                        self.model.pp_version + 1,
                        self.model.pp_version,
                    )
                    self.model.pp_version += 1

            if self.secondary_queue.qsize() >= self.reduce_batch_size:
                # Perform reduction on the secondary reduction queue
                # All processes(including master) in the reduction
                # group will execute this branch.
                grad_dict = self._reduce_batch(
                    self.secondary_queue,
                    self.reduce_batch_size,
                    self.reduce_method,
                    self.reduce_device,
                )
                # Push reduced results to the master queue.
                self.group.registered_sync(
                    self.primary_service, args=(grad_dict, ReduceType.REDUCE_PRIMARY)
                )

    @staticmethod
    def _reduce_batch(queue, batch_size, reduce_method, reduce_device):
        """
        Perform batched gradient reduction

        Returns:
            Reduced gradient dictionary.
        """
        batch = []
        while len(batch) < batch_size:
            batch.append(queue.get())
        grad_dict = {}
        for grad in batch:
            for k, v in grad.items():
                if k not in grad_dict:
                    grad_dict[k] = [v.to(reduce_device)]
                else:
                    grad_dict[k].append(v.to(reduce_device))
        for k, v in grad_dict.items():
            # Stack parameter tensors in dim 0 and reduce.
            if reduce_method == "sum":
                grad_dict[k] = t.sum(t.stack(v, dim=0), dim=0, keepdim=False)
            elif reduce_method == "mean":
                grad_dict[k] = t.mean(t.stack(v, dim=0), dim=0, keepdim=False)
            else:  # pragma: no cover
                raise RuntimeError("Unknown reduce method.")
        return grad_dict



from os.path import join
from typing import Dict, Any, Callable, Union
from torchviz import make_dot
import torch as t

from machin.utils.conf import Config
from machin.utils.prepare import prep_load_model
from machin.utils.logging import default_logger


class TorchFramework:
    """
    Base framework for all algorithms
    """

    _is_top = []
    _is_restorable = []

    def __init__(self):
        self._visualized = set()
        self._backward = t.autograd.backward
        # may be used by some distributed frameworks
        self.role = None

    @property
    def optimizers(self):
        raise NotImplementedError

    @optimizers.setter
    def optimizers(self, optimizers):
        raise NotImplementedError

    @property
    def lr_schedulers(self):
        raise NotImplementedError

    @property
    def top_models(self):
        models = []
        for m in self._is_top:
            models.append(getattr(self, m))
        return models

    @property
    def restorable_models(self):
        models = []
        for m in self._is_restorable:
            models.append(getattr(self, m))
        return models

    @property
    def backward_function(self):
        return self._backward

    @classmethod
    def get_top_model_names(cls):
        """
        Get attribute name of top level nn models.
        """
        return cls._is_top

    @classmethod
    def get_restorable_model_names(cls):
        """
        Get attribute name of restorable nn models.
        """
        return cls._is_restorable

    @classmethod
    def is_distributed(cls):
        """
        Whether this framework is a distributed framework which require
        multiple processes to run, and depends on ``torch.distributed`` or
        ``torch.distributed.rpc``
        """
        return False

    def set_backward_function(self, backward_func: Callable):
        """
        Replace the default backward function with a custom function.
        The default loss backward function is ``torch.autograd.backward``
        """
        assert callable(backward_func), "Backward function must be callable."
        self._backward = backward_func

    def enable_multiprocessing(self):
        """
        Enable multiprocessing for all modules.
        """
        for top in self._is_top:
            model = getattr(self, top)
            model.share_memory()

    def load(
        self, model_dir: str, network_map: Dict[str, str] = None, version: int = -1
    ):
        """
        Load models.

        An example of network map::

            {"restorable_model_1": "file_name_1",
             "restorable_model_2": "file_name_2"}

        Get keys by calling ``<Class name>.get_restorable()``

        Args:
            model_dir: Save directory.
            network_map: Key is module name, value is saved name.
            version: Version number of the save to be loaded.
        """
        network_map = {} if network_map is None else network_map
        restore_map = {}
        for r in self._is_restorable:
            if r in network_map:
                restore_map[network_map[r]] = getattr(self, r)
            else:
                default_logger.warning(
                    f'Load path for module "{r}" is not specified, module name is used.'
                )
                restore_map[r] = getattr(self, r)
        prep_load_model(model_dir, restore_map, version)

    def save(
        self, model_dir: str, network_map: Dict[str, str] = None, version: int = 0
    ):
        """
        Save models.

        An example of network map::

            {"restorable_model_1": "file_name_1",
             "restorable_model_2": "file_name_2"}

        Get keys by calling ``<Class name>.get_restorable()``

        Args:
            model_dir: Save directory.
            network_map: Key is module name, value is saved name.
            version: Version number of the new save.
        """
        network_map = {} if network_map is None else network_map
        if version == -1:
            version = "default"
            default_logger.warning(
                "You are using the default version to save, "
                "use custom version instead."
            )
        for r in self._is_restorable:
            if r in network_map:
                t.save(
                    getattr(self, r), join(model_dir, f"{network_map[r]}_{version}.pt"),
                )
            else:
                default_logger.warning(
                    'Save name for module "{r}" is not specified, module name is used.'
                )
                t.save(getattr(self, r), join(model_dir, f"{r}_{version}.pt"))

    def visualize_model(self, final_tensor: t.Tensor, name: str, directory: str):
        if name in self._visualized:
            return
        else:
            self._visualized.add(name)
            g = make_dot(final_tensor)
            g.render(
                filename=name,
                directory=directory,
                view=False,
                cleanup=False,
                quiet=True,
            )

    @classmethod
    def generate_config(cls, config: Union[Dict[str, Any], Config]):
        raise NotImplementedError

    @classmethod
    def init_from_config(
        cls,
        config: Union[Dict[str, Any], Config],
        model_device: Union[str, t.device] = "cpu",
    ):
        raise NotImplementedError
